import abc
import json
import requests

from typing import Optional, Union

from wtfis.models.base import WhoisBase
from wtfis.models.types import DomainEnrichmentType, IpEnrichmentType


class AbstractAttribute:
    def __get__(self, obj, type):  # pragma: no coverage
        raise NotImplementedError("This attribute must be set")


class BaseClient(abc.ABC):
    """
    Base client
    All clients should at least inherit from this class
    """
    @property
    @abc.abstractmethod
    def name(self) -> str:  # pragma: no coverage
        return NotImplemented


class BaseRequestsClient(BaseClient):
    """
    Client that uses the requests library
    """
    baseurl: Union[AbstractAttribute, str] = AbstractAttribute()

    def __init__(self) -> None:
        self.s = requests.Session()

    def _get(
        self,
        request: str,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
    ) -> dict:
        resp = self.s.get(self.baseurl + request, params=params, headers=headers)
        resp.raise_for_status()

        return json.loads(json.dumps((resp.json())))

    def _post(
        self,
        request: str,
        data: Optional[dict] = None,
        headers: Optional[dict] = None,
    ) -> dict:
        resp = self.s.post(self.baseurl + request, data=data, headers=headers)
        resp.raise_for_status()

        return json.loads(json.dumps((resp.json())))


class BaseWhoisClient(abc.ABC):
    """
    Client used for whois lookups
    """
    @abc.abstractmethod
    def get_whois(self, entity: str) -> WhoisBase:  # pragma: no coverage
        return NotImplemented


class BaseDomainEnricherClient(abc.ABC):
    """
    Client used for domain/FQDN enrichments
    """
    @abc.abstractmethod
    def enrich_domains(self, *domains: str) -> DomainEnrichmentType:  # pragma: no coverage
        return NotImplemented


class BaseIpEnricherClient(abc.ABC):
    """
    Client used for IP enrichments
    """
    @abc.abstractmethod
    def enrich_ips(self, *ips: str) -> IpEnrichmentType:  # pragma: no coverage
        return NotImplemented



from rich.console import Console
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TaskProgressColumn,
    TimeElapsedColumn,
    TextColumn,
)


def get_progress(console: Console) -> Progress:
    return Progress(
        SpinnerColumn(finished_text="[green]✓"),
        TextColumn("[bold]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        TimeElapsedColumn(),
        console=console,
    )



from __future__ import annotations

from typing import Optional, List, Dict

from pydantic import BaseModel, Field, RootModel


class AbuseIpDb(BaseModel):
    ip_address: str = Field(alias="ipAddress")
    is_public: Optional[bool] = Field(None, alias="isPublic")
    ip_version: Optional[int] = Field(None, alias="ipVersion")
    is_whitelisted: Optional[bool] = Field(None, alias="isWhitelisted")
    abuse_confidence_score: int = Field(alias="abuseConfidenceScore")
    country_code: Optional[str] = Field(None, alias="countryCode")
    usage_type: Optional[str] = Field(None, alias="usageType")
    isp: str
    domain: Optional[str] = None
    hostnames: Optional[List[str]] = None
    is_tor: Optional[bool] = Field(None, alias="isTor")
    total_reports: Optional[int] = Field(None, alias="totalReports")
    num_distinct_users: Optional[int] = Field(None, alias="numDistinctUsers")
    last_reported_at: Optional[str] = Field(None, alias="lastReportedAt")


class AbuseIpDbMap(RootModel):
    root: Dict[str, AbuseIpDb]

    @classmethod
    def empty(cls) -> AbuseIpDbMap:
        return cls.model_validate({})



import json
from wtfis.models.virustotal import (
    Domain,
    Resolutions,
)


class TestVirustotalModels:
    def test_domain_1(self, test_data):
        domain = Domain.model_validate(json.loads(test_data("vt_domain_gist.json")))

        assert domain.data.id_ == "gist.github.com"
        assert domain.data.type_ == "domain"
        assert domain.data.attributes.last_analysis_stats.malicious == 0
        assert domain.data.attributes.last_analysis_stats.suspicious == 0
        assert domain.data.attributes.last_analysis_stats.harmless == 84
        assert domain.data.attributes.last_analysis_stats.undetected == 10
        assert list(domain.data.attributes.last_analysis_results.root.keys())[:2] == [
            "CMC Threat Intelligence",
            "Snort IP sample list",
        ]

    def test_resolutions_1(self, test_data):
        res = Resolutions.model_validate(json.loads(test_data("vt_resolutions_gist.json")))

        assert res.meta.count == 37
        assert len(res.data) == 10
        assert res.data[1].attributes.ip_address == "192.30.255.113"
        assert res.data[1].attributes.ip_address_last_analysis_stats.malicious == 1
        assert res.data[1].attributes.ip_address_last_analysis_stats.suspicious == 0
        assert res.data[1].attributes.ip_address_last_analysis_stats.harmless == 82
        assert res.data[1].attributes.ip_address_last_analysis_stats.undetected == 11
        assert res.data[1].attributes.date == 1655835054



import pytest
from pathlib import Path
from typing import Optional

from rich.console import RenderableType
from rich.text import Span, Text

from wtfis.models.abuseipdb import AbuseIpDb
from wtfis.models.greynoise import GreynoiseIp
from wtfis.models.ipwhois import IpWhois
from wtfis.models.shodan import ShodanIp
from wtfis.models.urlhaus import UrlHaus


class TestTheme:
    """ Expected theme values for the tests """
    panel_title = "bold yellow"
    heading_h1 = "bold bright_green on dark_green"
    heading_h2 = "bold yellow"
    table_field = "bold bright_magenta"
    table_value = "not bold default"
    popularity_source = "bright_cyan"
    inline_stat = "cyan"
    vendor_list = "cyan"
    nameserver_list = "bright_blue"
    disclaimer = "italic white on red"
    tags = "bright_white on black"
    tags_green = "bright_green on black"
    tags_red = "bright_red on black"
    product = "orange_red1"
    port = "bright_cyan"
    transport = "cyan"
    footer = "cyan"
    info = "bold green"
    warn = "bold yellow"
    error = "bold red"
    timestamp_date = "not bold default"
    timestamp_t = "dim bright_white"
    timestamp_time = "dim default"
    timestamp_z = "dim bright_white"
    asn_org = "bright_white"
    whois_org = "bright_cyan"
    urlhaus_bl_name = "bright_cyan"
    urlhaus_bl_low = "bright_white on green"
    urlhaus_bl_med = "black on yellow"
    urlhaus_bl_high = "bright_white on red"


def open_test_data(fname: str) -> str:
    path = Path(__file__).parent.resolve() / "test_data" / fname
    with open(path) as f:
        return f.read()


def abuseipdb_get_ip(ip, pool) -> AbuseIpDb:
    """ Mock replacement for AbuseIpDbClient()._get_ip() """
    return AbuseIpDb.model_validate(pool[ip])


def greynoise_get(ip, pool) -> GreynoiseIp:
    """ Mock replacement for GreynoiseClient().get_ip() """
    return GreynoiseIp.model_validate(pool[ip])


def ipwhois_get(ip, pool) -> IpWhois:
    """ Mock replacement for IpWhoisClient().get_ipwhois() """
    return IpWhois.model_validate(pool[ip])


def shodan_get_ip(ip, pool) -> ShodanIp:
    """ Mock replacement for ShodanClient().get_ip() """
    return ShodanIp.model_validate(pool[ip])


def urlhaus_get_host(entity, pool) -> UrlHaus:
    """ Mock replacement for UrlHausClient()._get_host() """
    return UrlHaus.model_validate(pool[entity])


def timestamp_text(ts) -> Optional[RenderableType]:
    """ Standard timestamp formatting """
    theme = TestTheme()
    return Text(
        ts,
        spans=[
            Span(10, 11, theme.timestamp_t),
            Span(11, 19, theme.timestamp_time),
            Span(19, 20, theme.timestamp_z),
        ]
    )


@pytest.fixture(scope="module")
def test_data():
    return open_test_data


@pytest.fixture(scope="module")
def theme():
    return TestTheme()


@pytest.fixture(scope="module")
def mock_abuseipdb_get():
    return abuseipdb_get_ip


@pytest.fixture(scope="module")
def mock_greynoise_get():
    return greynoise_get


@pytest.fixture(scope="module")
def mock_ipwhois_get():
    return ipwhois_get


@pytest.fixture(scope="module")
def mock_shodan_get_ip():
    return shodan_get_ip


@pytest.fixture(scope="module")
def mock_urlhaus_get():
    return urlhaus_get_host


@pytest.fixture(scope="module")
def display_timestamp():
    return timestamp_text



from __future__ import annotations

from collections import defaultdict, namedtuple
from pydantic import BaseModel, RootModel
from typing import Dict, List, Optional

from wtfis.models.base import LaxStr


class PortData(BaseModel):
    port: int
    product: Optional[str] = None
    transport: str


class ShodanIp(BaseModel):
    asn: Optional[LaxStr] = None
    city: Optional[str] = None
    country_code: str
    country_name: str
    data: List[PortData]
    ip_str: str
    isp: Optional[str] = None
    last_update: str
    org: Optional[str] = None
    os: Optional[str] = None
    ports: Optional[List[int]] = None
    region_name: Optional[str] = None
    tags: Optional[List[str]] = None

    def group_ports_by_product(self) -> dict:
        PortProtocol = namedtuple("PortProtocol", "port transport")
        result = defaultdict(list)
        unknown = []  # Save ports with no product for adding later as last result item
        for port in self.data:
            port_data = PortProtocol(port.port, port.transport)
            if port.product:
                result[port.product].append(port_data)
            else:
                unknown.append(port_data)
        if unknown:
            result["Other"] = unknown
        return result


class ShodanIpMap(RootModel):
    root: Dict[str, ShodanIp]

    @classmethod
    def empty(cls) -> ShodanIpMap:
        return cls.model_validate({})



import json
from unittest.mock import MagicMock, patch

import pytest
from requests.exceptions import ConnectionError
from rich.console import Console

from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.base import requests
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.passivetotal import PTClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient
from wtfis.handlers.domain import DomainHandler
from wtfis.handlers.ip import IpAddressHandler
from wtfis.models.greynoise import GreynoiseIpMap
from wtfis.models.ipwhois import IpWhoisMap
from wtfis.models.virustotal import Resolutions


def generate_domain_handler(max_resolutions=3):
    return DomainHandler(
        entity="www.example[.]com",
        console=Console(),
        progress=MagicMock(),
        vt_client=VTClient("dummykey"),
        ip_geoasn_client=IpWhoisClient(),
        whois_client=PTClient("dummyuser", "dummykey"),
        shodan_client=ShodanClient("dummykey"),
        greynoise_client=GreynoiseClient("dummykey"),
        abuseipdb_client=AbuseIpDbClient("dummykey"),
        urlhaus_client=UrlHausClient(),
        max_resolutions=max_resolutions,
    )


def generate_ip_handler():
    return IpAddressHandler(
        entity="1[.]1[.]1[.]1",
        console=Console(),
        progress=MagicMock(),
        vt_client=VTClient("dummykey"),
        ip_geoasn_client=IpWhoisClient(),
        whois_client=PTClient("dummyuser", "dummykey"),
        shodan_client=ShodanClient("dummykey"),
        greynoise_client=GreynoiseClient("dummykey"),
        abuseipdb_client=AbuseIpDbClient("dummykey"),
        urlhaus_client=UrlHausClient(),
    )


@pytest.fixture
def domain_handler():
    return generate_domain_handler


@pytest.fixture
def ip_handler():
    return generate_ip_handler


class TestDomainHandler:
    def test_entity_refang(self, domain_handler):
        handler = domain_handler()
        assert handler.entity == "www.example.com"

    def test_fetch_data_1(self, domain_handler, test_data):
        """ Test with max_resolutions = 3 (default) """
        handler = domain_handler()
        handler.resolutions = Resolutions.model_validate(json.loads(test_data("vt_resolutions_gist.json")))

        handler._fetch_vt_domain = MagicMock()
        handler._fetch_vt_resolutions = MagicMock()
        handler._fetch_geoasn = MagicMock()
        handler._fetch_whois = MagicMock()
        handler._fetch_shodan = MagicMock()
        handler._fetch_greynoise = MagicMock()
        handler._fetch_urlhaus = MagicMock()

        handler.fetch_data()
        handler._fetch_vt_domain.assert_called_once()
        handler._fetch_vt_resolutions.assert_called_once()
        handler._fetch_geoasn.assert_called_once()
        handler._fetch_whois.assert_called_once()
        handler._fetch_shodan.assert_called_once()
        handler._fetch_greynoise.assert_called_once()
        handler._fetch_urlhaus.assert_called_once()

    def test_fetch_data_2(self, domain_handler):
        """ Test with max_resolutions = 0 """
        handler = domain_handler(0)
        handler._fetch_vt_domain = MagicMock()
        handler._fetch_vt_resolutions = MagicMock()
        handler._fetch_geoasn = MagicMock()
        handler._fetch_whois = MagicMock()
        handler._fetch_shodan = MagicMock()
        handler._fetch_greynoise = MagicMock()
        handler._fetch_urlhaus = MagicMock()

        handler.fetch_data()
        handler._fetch_vt_domain.assert_called_once()
        handler._fetch_vt_resolutions.assert_not_called()
        handler._fetch_geoasn.assert_not_called()
        handler._fetch_whois.assert_called_once()
        handler._fetch_shodan.assert_not_called()
        handler._fetch_greynoise.assert_not_called()
        handler._fetch_urlhaus.assert_called_once()

        assert handler.geoasn == IpWhoisMap.model_validate({})
        assert handler.geoasn.root == {}

    @patch.object(requests.Session, "get")
    def test_vt_http_error(self, mock_requests_get, domain_handler, capsys):
        """
        Test a requests HTTPError from the VT client. This also tests the
        common_exception_handler decorator.
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 401
        mock_requests_get.return_value = mock_resp

        # Thorough test of first _fetch_* method
        with pytest.raises(SystemExit) as e:
            handler._fetch_vt_domain()

        capture = capsys.readouterr()

        assert capture.err == "Error fetching data: 401 Client Error: None for url: None\n"
        assert e.type == SystemExit
        assert e.value.code == 1

        # Extra: just make sure program exits correctly
        with pytest.raises(SystemExit) as e:
            handler._fetch_vt_resolutions()
        assert e.value.code == 1

    @patch.object(requests.Session, "get")
    def test_vt_validation_error(self, mock_requests_get, domain_handler, capsys):
        """
        Test a pydantic data model ValidationError from the VT client. This also tests the
        common_exception_handler decorator.
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            # Thorough test of first _fetch_* method
            with pytest.raises(SystemExit) as e:
                handler._fetch_vt_domain()

            capture = capsys.readouterr()

            assert capture.err.startswith(
                "Data model validation error: 1 validation error for Domain\ndata\n"
                "  Field required [type=missing, input_value={'intentionally': 'wrong data'}, input_type=dict]\n"
            )
            assert e.type == SystemExit
            assert e.value.code == 1

            # Extra: just make sure program exits correctly
            with pytest.raises(SystemExit) as e:
                handler._fetch_vt_resolutions()
            assert e.value.code == 1

    @patch.object(requests.Session, "get")
    def test_ipwhois_http_error(self, mock_requests_get, domain_handler, capsys, test_data):
        handler = domain_handler()
        handler.resolutions = Resolutions.model_validate(json.loads(test_data("vt_resolutions_gist.json")))

        mock_resp = requests.models.Response()

        mock_resp.status_code = 500
        mock_requests_get.return_value = mock_resp

        handler._fetch_geoasn(*["1.2.3.4", "1.2.3.5"])
        assert handler.warnings[0].startswith("Could not fetch IPWhois: 500 Server Error:")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch IPWhois: 500 Server Error:")

    @patch.object(requests.Session, "get")
    def test_ipwhois_validation_error(self, mock_requests_get, domain_handler, capsys):
        handler = domain_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"success": True, "intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            with pytest.raises(SystemExit) as e:
                handler._fetch_geoasn("1.2.3.4")

            capture = capsys.readouterr()

            assert capture.err.startswith(
                "Data model validation error: 9 validation errors for IpWhois\n"
            )
            assert e.type == SystemExit
            assert e.value.code == 1

    @patch.object(requests.Session, "get")
    def test_whois_http_error(self, mock_requests_get, domain_handler, capsys):
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 401
        mock_requests_get.return_value = mock_resp

        with pytest.raises(SystemExit) as e:
            handler._fetch_whois()

        capture = capsys.readouterr()

        assert capture.err == "Error fetching data: 401 Client Error: None for url: None\n"
        assert e.type == SystemExit
        assert e.value.code == 1

    @patch.object(requests.Session, "get")
    def test_vt_resolutions_429_error(self, mock_requests_get, domain_handler, capsys):
        """
        Test fail open behavior of VT resolution fetching when rate limited
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler._fetch_vt_resolutions()
        assert handler.warnings[0].startswith("Could not fetch Virustotal resolutions: 429 Client Error:")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Virustotal resolutions: 429 Client Error:")

    @patch.object(requests.Session, "get")
    def test_whois_429_error(self, mock_requests_get, domain_handler, capsys):
        """
        Test fail open behavior of whois fetching when rate limited
        Since _fetch_whois() is in the base class, this should also cover the IpAddressHandler use case.
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler._fetch_whois()
        assert handler.warnings[0].startswith("Could not fetch Whois: 429 Client Error:")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Whois: 429 Client Error:")

    @patch.object(requests.Session, "get")
    def test_greynoise_429_error(self, mock_requests_get, domain_handler, capsys, test_data):
        """
        Test fail open behavior of Greynoise when rate limited
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler.resolutions = Resolutions.model_validate(json.loads(test_data("vt_resolutions_gist.json")))

        handler._fetch_greynoise(*["1.2.3.4", "1.2.3.5"])
        assert handler.warnings[0].startswith("Could not fetch Greynoise: 429 Client Error:")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Greynoise: 429 Client Error:")

    @patch.object(requests.Session, "get")
    def test_greynoise_404_error(self, mock_requests_get, domain_handler, test_data):
        """
        Test fail open behavior of Greynoise when no IP found (404) and no warning message
        """
        handler = domain_handler(3)
        mock_resp = requests.models.Response()

        handler.resolutions = Resolutions.model_validate(json.loads(test_data("vt_resolutions_gist.json")))

        mock_resp.status_code = 404
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise(*["1.2.3.4", "1.2.3.5"])
        assert len(handler.warnings) == 0
        assert handler.greynoise == GreynoiseIpMap.model_validate({})

    @patch.object(requests.Session, "get")
    def test_greynoise_403_error(self, mock_requests_get, domain_handler, test_data, capsys):
        """
        Test exception behavior of Greynoise when not under any of the handled cases
        """
        handler = domain_handler(3)
        mock_resp = requests.models.Response()

        handler.resolutions = Resolutions.model_validate(json.loads(test_data("vt_resolutions_gist.json")))

        mock_resp.status_code = 403
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise(*["1.2.3.4", "1.2.3.5"])
        assert handler.warnings[0].startswith("Could not fetch Greynoise: 403 Client Error:")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Greynoise: 403 Client Error:")

    @patch.object(requests.Session, "get")
    def test_greynoise_connection_error(self, mock_requests_get, domain_handler, test_data, capsys):
        """
        Test exception behavior of Greynoise with non-HTTPError requests exception
        """
        handler = domain_handler(3)
        mock_requests_get.side_effect = ConnectionError("Foo bar message")

        handler.resolutions = Resolutions.model_validate(json.loads(test_data("vt_resolutions_gist.json")))

        handler._fetch_greynoise(*["1.2.3.4", "1.2.3.5"])
        assert handler.warnings[0] == "Could not fetch Greynoise: Foo bar message"

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Greynoise: Foo bar message")

    @patch.object(requests.Session, "get")
    def test_greynoise_validation_error(self, mock_requests_get, domain_handler, capsys):
        handler = domain_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            with pytest.raises(SystemExit) as e:
                handler._fetch_greynoise("1.2.3.4")

            capture = capsys.readouterr()

            assert capture.err.startswith(
                "Data model validation error: 5 validation errors for GreynoiseIp\n"
            )
            assert e.type == SystemExit
            assert e.value.code == 1

    @patch.object(requests.Session, "post")
    def test_urlhaus_http_error(self, mock_requests_post, domain_handler, capsys):
        """
        Test exception behavior of URLhaus with non-HTTPError requests exception
        """
        handler = domain_handler()
        mock_requests_post.side_effect = ConnectionError("Foo bar message")

        handler._fetch_urlhaus()
        assert handler.warnings[0] == "Could not fetch URLhaus: Foo bar message"

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch URLhaus: Foo bar message")


class TestIpAddressHandler:
    def test_entity_refang(self, ip_handler):
        handler = ip_handler()
        assert handler.entity == "1.1.1.1"

    def test_fetch_data(self, ip_handler):
        handler = ip_handler()
        handler._fetch_vt_ip_address = MagicMock()
        handler._fetch_geoasn = MagicMock()
        handler._fetch_whois = MagicMock()
        handler._fetch_shodan = MagicMock()
        handler._fetch_greynoise = MagicMock()
        handler._fetch_urlhaus = MagicMock()
        handler._fetch_abuseipdb = MagicMock()

        handler.fetch_data()
        handler._fetch_vt_ip_address.assert_called_once()
        handler._fetch_geoasn.assert_called_once()
        handler._fetch_whois.assert_called_once()
        handler._fetch_shodan.assert_called_once()
        handler._fetch_greynoise.assert_called_once()
        handler._fetch_urlhaus.assert_called_once()
        handler._fetch_abuseipdb.assert_called_once()

    @patch.object(requests.Session, "get")
    def test_vt_http_error(self, mock_requests_get, ip_handler, capsys):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 404
        mock_requests_get.return_value = mock_resp

        # Thorough test of first _fetch_* method
        with pytest.raises(SystemExit) as e:
            handler._fetch_vt_ip_address()

        capture = capsys.readouterr()

        assert capture.err == "Error fetching data: 404 Client Error: None for url: None\n"
        assert e.type == SystemExit
        assert e.value.code == 1

    @patch.object(requests.Session, "get")
    def test_vt_validation_error(self, mock_requests_get, ip_handler, capsys):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            # Thorough test of first _fetch_* method
            with pytest.raises(SystemExit) as e:
                handler._fetch_vt_ip_address()

            capture = capsys.readouterr()

            assert capture.err.startswith(
                "Data model validation error: 1 validation error for IpAddress\ndata\n"
                "  Field required [type=missing, input_value={'intentionally': 'wrong data'}, input_type=dict]\n"
            )
            assert e.type == SystemExit
            assert e.value.code == 1

    @patch.object(requests.Session, "get")
    def test_ipwhois_http_error(self, mock_requests_get, ip_handler, capsys):
        handler = ip_handler()

        mock_resp = requests.models.Response()

        mock_resp.status_code = 502
        mock_requests_get.return_value = mock_resp

        handler._fetch_geoasn("1.2.3.4")
        assert handler.warnings[0].startswith("Could not fetch IPWhois: 502 Server Error:")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch IPWhois: 502 Server Error:")

    @patch.object(requests.Session, "get")
    def test_whois_http_error(self, mock_requests_get, ip_handler, capsys):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 403
        mock_requests_get.return_value = mock_resp

        with pytest.raises(SystemExit) as e:
            handler._fetch_whois()

        capture = capsys.readouterr()

        assert capture.err == "Error fetching data: 403 Client Error: None for url: None\n"
        assert e.type == SystemExit
        assert e.value.code == 1

    @patch.object(requests.Session, "get")
    def test_ipwhois_validation_error(self, mock_requests_get, ip_handler, capsys):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"success": True, "intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            with pytest.raises(SystemExit) as e:
                handler._fetch_geoasn("1.2.3.4")

            capture = capsys.readouterr()

            assert capture.err.startswith(
                "Data model validation error: 9 validation errors for IpWhois\n"
            )
            assert e.type == SystemExit
            assert e.value.code == 1

    @patch.object(requests.Session, "get")
    def test_shodan_api_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test fail open behavior of Shodan on invalid API key
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 401
        mock_resp._content = b'<html>'
        mock_requests_get.return_value = mock_resp

        handler._fetch_shodan("1.2.3.4")
        assert handler.warnings[0].startswith("Could not fetch Shodan: Invalid API key")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Shodan: Invalid API key")

    @patch.object(requests.Session, "get")
    def test_greynoise_http_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test Greynoise HTTP error that results in a SystemExit
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 401
        mock_requests_get.return_value = mock_resp

        with pytest.raises(SystemExit) as e:
            handler._fetch_whois()

        capture = capsys.readouterr()

        assert capture.err == "Error fetching data: 401 Client Error: None for url: None\n"
        assert e.type == SystemExit
        assert e.value.code == 1

    @patch.object(requests.Session, "get")
    def test_greynoise_429_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test fail open behavior of Greynoise when rate limited
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise("1.2.3.4")
        assert handler.warnings[0].startswith("Could not fetch Greynoise: 429 Client Error:")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Greynoise: 429 Client Error:")

    @patch.object(requests.Session, "get")
    def test_greynoise_404_error(self, mock_requests_get, ip_handler):
        """
        Test fail open behavior of Greynoise when no IP found (404) and no warning message
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 404
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise()
        assert len(handler.warnings) == 0
        assert handler.greynoise == GreynoiseIpMap.model_validate({})

    @patch.object(requests.Session, "get")
    def test_greynoise_403_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test exception behavior of Greynoise when not under any of the handled cases
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 403
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise("1.2.3.4")
        assert handler.warnings[0].startswith("Could not fetch Greynoise: 403 Client Error:")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Greynoise: 403 Client Error:")

    @patch.object(requests.Session, "get")
    def test_greynoise_connection_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test exception behavior of Greynoise with non-HTTPError requests exception
        """
        handler = ip_handler()
        mock_requests_get.side_effect = ConnectionError("Foo bar message")

        handler._fetch_greynoise("1.2.3.4")
        assert handler.warnings[0] == "Could not fetch Greynoise: Foo bar message"

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Greynoise: Foo bar message")

    @patch.object(requests.Session, "get")
    def test_greynoise_validation_error(self, mock_requests_get, ip_handler, capsys):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            with pytest.raises(SystemExit) as e:
                handler._fetch_greynoise("1.2.3.4")

            capture = capsys.readouterr()

            assert capture.err.startswith(
                "Data model validation error: 5 validation errors for GreynoiseIp\n"
            )
            assert e.type == SystemExit
            assert e.value.code == 1

    @patch.object(requests.Session, "post")
    def test_urlhaus_http_error(self, mock_requests_post, ip_handler, capsys):
        """
        Test exception behavior of URLhaus with non-HTTPError requests exception
        """
        handler = ip_handler()
        mock_requests_post.side_effect = ConnectionError("Foo bar message")

        handler._fetch_urlhaus()
        assert handler.warnings[0] == "Could not fetch URLhaus: Foo bar message"

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch URLhaus: Foo bar message")

    @patch.object(requests.Session, "get")
    def test_abuseipdb_429_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test fail open behavior of AbuseIPDB when rate limited
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler._fetch_abuseipdb("1.2.3.4")
        assert handler.warnings[0].startswith("Could not fetch AbuseIPDB: 429 Client Error:")

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch AbuseIPDB: 429 Client Error:")



from wtfis.clients.base import BaseRequestsClient, BaseWhoisClient
from wtfis.models.virustotal import (
    Domain,
    IpAddress,
    Resolutions,
    Whois,
)
from wtfis.utils import is_ip


class VTClient(BaseRequestsClient, BaseWhoisClient):
    """
    Virustotal client
    """
    baseurl = "https://www.virustotal.com/api/v3"

    def __init__(self, api_key: str) -> None:
        super().__init__()
        self.s.headers.update({
            "x-apikey": api_key,
            "Accept": "application/json",
        })

    @property
    def name(self) -> str:
        return "Virustotal"

    def get_domain(self, domain: str) -> Domain:
        return Domain.model_validate(self._get(f"/domains/{domain}"))

    def get_domain_resolutions(self, domain: str) -> Resolutions:
        return Resolutions.model_validate(self._get(f"/domains/{domain}/resolutions"))

    def get_domain_whois(self, domain: str) -> Whois:
        return Whois.model_validate(self._get(f"/domains/{domain}/historical_whois"))

    def get_ip_address(self, ip: str) -> IpAddress:
        return IpAddress.model_validate(self._get(f"/ip_addresses/{ip}"))

    def get_ip_whois(self, ip: str) -> Whois:
        return Whois.model_validate(self._get(f"/ip_addresses/{ip}/historical_whois"))

    def get_whois(self, entity: str) -> Whois:
        """ Generalized for domain and IP """
        if is_ip(entity):
            return self.get_ip_whois(entity)
        else:
            return self.get_domain_whois(entity)



from rich.columns import Columns
from rich.console import (
    Console,
    Group,
    RenderableType,
)
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text
from typing import List, Optional, Tuple, Union
from wtfis.models.abuseipdb import AbuseIpDbMap

from wtfis.models.base import WhoisBase
from wtfis.models.greynoise import GreynoiseIpMap
from wtfis.models.shodan import ShodanIpMap
from wtfis.models.urlhaus import UrlHausMap
from wtfis.models.types import IpGeoAsnMapType
from wtfis.models.virustotal import (
    Domain,
    IpAddress,
    Resolutions,
)
from wtfis.ui.base import BaseView
from wtfis.utils import Timestamp, smart_join


class DomainView(BaseView):
    """
    Handler for FQDN and domain lookup output
    """

    def __init__(
        self,
        console: Console,
        entity: Domain,
        resolutions: Optional[Resolutions],
        geoasn: IpGeoAsnMapType,
        whois: WhoisBase,
        shodan: ShodanIpMap,
        greynoise: GreynoiseIpMap,
        abuseipdb: AbuseIpDbMap,
        urlhaus: UrlHausMap,
        max_resolutions: int = 3,
    ) -> None:
        super().__init__(console, entity, geoasn, whois, shodan, greynoise, abuseipdb, urlhaus)

        self.resolutions = resolutions
        self.max_resolutions = max_resolutions

    def domain_panel(self) -> Panel:
        content = [self._gen_vt_section()]  # VT section
        for section in (
            self._gen_urlhaus_section(),    # URLhaus section
        ):
            if section is not None:
                content.append("")
                content.append(section)

        return self._gen_panel(self._gen_group(content), self.entity.data.id_)

    def resolutions_panel(self) -> Optional[Panel]:
        # Skip if no resolutions data
        if not self.resolutions:
            return None

        content: List[RenderableType] = [self._gen_heading_text("Resolutions")]

        for idx, ip in enumerate(self.resolutions.data):
            if idx == self.max_resolutions:
                break
            attributes = ip.attributes

            # Analysis
            analysis = self._gen_vt_analysis_stats(attributes.ip_address_last_analysis_stats)
            analysis_field = self._gen_linked_field_name(
                "Analysis",
                hyperlink=f"{self.vt_gui_baseurl_ip}/{attributes.ip_address}",
            )

            # Content
            heading = self._gen_heading_text(
                attributes.ip_address,
                # hyperlink=f"{self.vt_gui_baseurl_ip}/{attributes.ip_address}",
                type="h2",
            )
            data: List[Tuple[Union[str, Text], Union[RenderableType, None]]] = [
                (analysis_field, analysis),
                ("Resolved:", Timestamp(attributes.date).render),
            ]

            # IP geolocation and ASN
            geoasn = self._get_geoasn_enrichment(attributes.ip_address)
            if geoasn:
                asn = self._gen_asn_text(geoasn.asn, geoasn.org)
                data += [
                    ("ASN:", asn),
                    ("ISP:", geoasn.isp),
                    ("Location:", smart_join(geoasn.city, geoasn.region, geoasn.country)),
                ]

            # Shodan
            shodan = self._get_shodan_enrichment(attributes.ip_address)
            if shodan:
                tags = smart_join(*shodan.tags, style=self.theme.tags) if shodan.tags else None
                services_field = self._gen_linked_field_name(
                    "Services",
                    hyperlink=f"{self.shodan_gui_baseurl}/{attributes.ip_address}"
                )
                data += [
                    ("OS:", shodan.os),
                    (services_field, self._gen_shodan_services(shodan)),
                    ("Tags:", tags),
                    # ("Last Scan:", Timestamp(f"{shodan.last_update}+00:00").render),  # Timestamps are UTC
                    #                                                                   # (source: Google)
                ]

            # Greynoise
            greynoise = self._get_greynoise_enrichment(attributes.ip_address)
            if greynoise:
                data += [self._gen_greynoise_tuple(greynoise)]

            # AbuseIPDB
            abuseipdb = self._get_abuseipdb_enrichment(attributes.ip_address)
            if abuseipdb:
                data += [self._gen_abuseipdb_tuple(abuseipdb)]

            # Include a disclaimer if last seen is older than 1 year
            # Note: Disabled for now because I originally understood that the resolution date was the last time
            # the domain was resolved, but it may actually be he first time the IP itself was seen with the domain.
            # if enrich and older_than(attributes.date, 365):
            #     body = Group(
            #         self._gen_table(*data),
            #         Text("**Enrichment data may be inaccurate", style=self.theme.disclaimer),
            #     )  # type: Union[Group, Table, str]
            # else:
            #     body = self._gen_table(*data)

            body = self._gen_table(*data)
            content.append(self._gen_section(body, heading))

            # Add extra line break if not last item in list
            if (
                idx < self.max_resolutions - 1 and
                idx < len(self.resolutions.data) - 1
            ):
                content.append("")

        # Info about how many more IPs were not shown
        if self.max_resolutions < self.resolutions.meta.count:
            content.append("")
            content.append(
                Text(justify="center", end="\n").append(
                    f"+{self.resolutions.meta.count - self.max_resolutions} more",
                    style=(f"{self.theme.footer} "
                           f"link {self.vt_gui_baseurl_domain}/{self.entity.data.id_}/relations"),
                )
            )

        # Render results, if existent
        if len(content) > 1:
            return self._gen_panel(self._gen_group(content))

        # No result
        return None

    def print(self, one_column: bool = False) -> None:
        renderables = [i for i in (
            self.domain_panel(),
            self.resolutions_panel(),
            self.whois_panel(),
        ) if i is not None]

        if one_column:
            self.console.print(Group(*([""] + renderables)))  # type: ignore
        else:
            self.console.print(Padding(Columns(renderables), (1, 0)))


class IpAddressView(BaseView):
    """
    Handler for IP Address lookup output
    """

    def __init__(
        self,
        console: Console,
        entity: IpAddress,
        geoasn: IpGeoAsnMapType,
        whois: WhoisBase,
        shodan: ShodanIpMap,
        greynoise: GreynoiseIpMap,
        abuseipdb: AbuseIpDbMap,
        urlhaus: UrlHausMap,
    ) -> None:
        super().__init__(console, entity, geoasn, whois, shodan, greynoise, abuseipdb, urlhaus)

    def ip_panel(self) -> Panel:
        content = [self._gen_vt_section()]  # VT section
        for section in (
            self._gen_geoasn_section(),     # IP location and ASN section
            self._gen_shodan_section(),     # Shodan section
            self._gen_urlhaus_section(),    # URLhaus section
            self._gen_ip_other_section(),   # Other section
        ):
            if section is not None:
                content.append("")
                content.append(section)

        return self._gen_panel(self._gen_group(content), self.entity.data.id_)

    def print(self, one_column: bool = False) -> None:
        renderables = [i for i in (
            self.ip_panel(),
            self.whois_panel(),
        ) if i is not None]

        if one_column:
            self.console.print(Group(*([""] + renderables)))  # type: ignore
        else:
            self.console.print(Padding(Columns(renderables), (1, 0)))



import argparse
import os

from argparse import Namespace
from dotenv import load_dotenv
from pathlib import Path
from rich.console import Console
from rich.progress import Progress
from typing import Union
from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.ip2whois import Ip2WhoisClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.passivetotal import PTClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient
from wtfis.handlers.base import BaseHandler
from wtfis.handlers.domain import DomainHandler
from wtfis.handlers.ip import IpAddressHandler
from wtfis.models.virustotal import Domain, IpAddress
from wtfis.utils import error_and_exit, is_ip
from wtfis.ui.base import BaseView
from wtfis.ui.progress import get_progress
from wtfis.ui.view import DomainView, IpAddressView
from wtfis.version import get_version


def parse_env() -> None:
    DEFAULT_ENV_FILE = Path().home() / ".env.wtfis"

    # Load the file
    load_dotenv(DEFAULT_ENV_FILE)

    # Exit if required environment variables don't exist
    for envvar in (
        "VT_API_KEY",
    ):
        if not os.environ.get(envvar):
            error = f"Error: Environment variable {envvar} not set"
            if not DEFAULT_ENV_FILE.exists():
                error = error + f"\nEnv file {DEFAULT_ENV_FILE} was not found either. Did you forget?"
            error_and_exit(error)


def parse_args() -> Namespace:
    DEFAULT_MAX_RESOLUTIONS = 3

    parser = argparse.ArgumentParser()
    parser.add_argument("entity", help="Hostname, domain or IP")
    parser.add_argument(
        "-m", "--max-resolutions", metavar="N",
        help=f"Maximum number of resolutions to show (default: {DEFAULT_MAX_RESOLUTIONS})",
        type=int,
        default=DEFAULT_MAX_RESOLUTIONS
    )
    parser.add_argument("-s", "--use-shodan", help="Use Shodan to enrich IPs", action="store_true")
    parser.add_argument("-g", "--use-greynoise", help="Enable Greynoise for IPs", action="store_true")
    parser.add_argument("-a", "--use-abuseipdb", help="Enable AbuseIPDB for IPs", action="store_true")
    parser.add_argument("-u", "--use-urlhaus", help="Enable URLhaus for IPs and domains", action="store_true")
    parser.add_argument("-n", "--no-color", help="Show output without colors", action="store_true")
    parser.add_argument("-1", "--one-column", help="Display results in one column", action="store_true")
    parser.add_argument(
        "-V", "--version",
        help="Print version number",
        action="version",
        version=get_version()
    )
    parsed = parser.parse_args()

    # Default overrides
    # If a default is set, then setting the flag as an argument _negates_ the effect
    for option in os.environ.get("WTFIS_DEFAULTS", "").split(" "):
        if option in ("-s", "--use-shodan"):
            parsed.use_shodan = not parsed.use_shodan
        elif option in ("-g", "--use-greynoise"):
            parsed.use_greynoise = not parsed.use_greynoise
        elif option in ("-a", "--use-abuseipdb"):
            parsed.use_abuseipdb = not parsed.use_abuseipdb
        elif option in ("-u", "--use-urlhaus"):
            parsed.use_urlhaus = not parsed.use_urlhaus
        elif option in ("-n", "--no-color"):
            parsed.no_color = not parsed.no_color
        elif option in ("-1", "--one-column"):
            parsed.one_column = not parsed.one_column

    # Validation
    if parsed.max_resolutions > 10:
        argparse.ArgumentParser().error("Maximum --max-resolutions value is 10")
    if parsed.use_shodan and not os.environ.get("SHODAN_API_KEY"):
        argparse.ArgumentParser().error("SHODAN_API_KEY is not set")
    if parsed.use_greynoise and not os.environ.get("GREYNOISE_API_KEY"):
        argparse.ArgumentParser().error("GREYNOISE_API_KEY is not set")
    if parsed.use_abuseipdb and not os.environ.get("ABUSEIPDB_API_KEY"):
        argparse.ArgumentParser().error("ABUSEIPDB_API_KEY is not set")
    if is_ip(parsed.entity) and parsed.max_resolutions != DEFAULT_MAX_RESOLUTIONS:
        argparse.ArgumentParser().error("--max-resolutions is not applicable to IPs")

    return parsed


def generate_entity_handler(
    args: Namespace,
    console: Console,
    progress: Progress,
) -> BaseHandler:
    # Virustotal client
    vt_client = VTClient(os.environ["VT_API_KEY"])

    # IP geolocation and ASN client selector
    # TODO: add more options
    ip_geoasn_client = IpWhoisClient()

    # Whois client selector
    # Order of use based on set envvars:
    #    1. Passivetotal
    #    2. IP2Whois (Domain only)
    #    2. Virustotal (fallback)
    if os.environ.get("PT_API_USER") and os.environ.get("PT_API_KEY"):
        whois_client: Union[PTClient, Ip2WhoisClient, VTClient] = (
            PTClient(os.environ["PT_API_USER"], os.environ["PT_API_KEY"])
        )
    elif os.environ.get("IP2WHOIS_API_KEY") and not is_ip(args.entity):
        whois_client = Ip2WhoisClient(os.environ["IP2WHOIS_API_KEY"])
    else:
        whois_client = vt_client

    shodan_client = (
        ShodanClient(os.environ["SHODAN_API_KEY"])
        if args.use_shodan
        else None
    )

    # Greynoise client (optional)
    greynoise_client = (
        GreynoiseClient(os.environ["GREYNOISE_API_KEY"])
        if args.use_greynoise
        else None
    )

    # AbuseIPDB client (optional)
    abuseipdb_client = (
        AbuseIpDbClient(os.environ["ABUSEIPDB_API_KEY"])
        if args.use_abuseipdb else None
    )

    # URLhaus client (optional)
    urlhaus_client = (
        UrlHausClient()
        if args.use_urlhaus
        else None
    )

    # Domain / FQDN handler
    if not is_ip(args.entity):
        entity: BaseHandler = DomainHandler(
            entity=args.entity,
            console=console,
            progress=progress,
            vt_client=vt_client,
            ip_geoasn_client=ip_geoasn_client,
            whois_client=whois_client,
            shodan_client=shodan_client,
            greynoise_client=greynoise_client,
            abuseipdb_client=abuseipdb_client,
            urlhaus_client=urlhaus_client,
            max_resolutions=args.max_resolutions,
        )
    # IP address handler
    else:
        entity = IpAddressHandler(
            entity=args.entity,
            console=console,
            progress=progress,
            vt_client=vt_client,
            ip_geoasn_client=ip_geoasn_client,
            whois_client=whois_client,
            shodan_client=shodan_client,
            greynoise_client=greynoise_client,
            abuseipdb_client=abuseipdb_client,
            urlhaus_client=urlhaus_client,
        )

    return entity


def generate_view(
    args: Namespace,
    console: Console,
    entity: BaseHandler,
) -> BaseView:
    # Output display
    if isinstance(entity, DomainHandler) and isinstance(entity.vt_info, Domain):
        view: BaseView = DomainView(
            console,
            entity.vt_info,
            entity.resolutions,
            entity.geoasn,
            entity.whois,
            entity.shodan,
            entity.greynoise,
            entity.abuseipdb,
            entity.urlhaus,
            max_resolutions=args.max_resolutions,
        )
    elif isinstance(entity, IpAddressHandler) and isinstance(entity.vt_info, IpAddress):
        view = IpAddressView(
            console,
            entity.vt_info,
            entity.geoasn,
            entity.whois,
            entity.shodan,
            entity.greynoise,
            entity.abuseipdb,
            entity.urlhaus,
        )
    else:
        raise Exception("Unsupported entity!")

    return view


def main():
    # Load environment variables
    parse_env()

    # Args
    args = parse_args()

    # Instantiate the console
    console = Console(no_color=True) if args.no_color else Console()

    # Progress animation controller
    progress = get_progress(console)

    # Entity handler
    entity = generate_entity_handler(args, console, progress)

    # Fetch data
    with progress:
        entity.fetch_data()

    # Print fetch warnings, if any
    entity.print_warnings()

    # Output display
    view = generate_view(args, console, entity)

    # Finally, print output
    view.print(one_column=args.one_column)



import pytest
import json

from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Span, Text
from unittest.mock import MagicMock
from wtfis.clients.abuseipdb import AbuseIpDbClient

from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.models.greynoise import GreynoiseIpMap
from wtfis.models.ipwhois import IpWhoisMap
from wtfis.models.passivetotal import Whois as PTWhois
from wtfis.models.urlhaus import UrlHausMap
from wtfis.models.virustotal import (
    IpAddress,
    Whois as VTWhois,
)
from wtfis.ui.view import IpAddressView


@pytest.fixture()
def view01(test_data, mock_abuseipdb_get, mock_ipwhois_get, mock_shodan_get_ip, mock_greynoise_get, mock_urlhaus_get):
    """ 1.1.1.1 with PT whois. Complete test of all panels. Also test print(). """
    ip = "1.1.1.1"

    geoasn_pool = json.loads(test_data("ipwhois_1.1.1.1.json"))
    geoasn_client = IpWhoisClient()
    geoasn_client._get_ipwhois = MagicMock(side_effect=lambda ip: mock_ipwhois_get(ip, geoasn_pool))
    geoasn_enrich = geoasn_client.enrich_ips(ip)

    shodan_pool = json.loads(test_data("shodan_1.1.1.1.json"))
    shodan_client = ShodanClient(MagicMock())
    shodan_client._get_ip = MagicMock(side_effect=lambda ip: mock_shodan_get_ip(ip, shodan_pool))
    shodan_enrich = shodan_client.enrich_ips(ip)

    greynoise_pool = json.loads(test_data("greynoise_1.1.1.1.json"))
    greynoise_client = GreynoiseClient("dummykey")
    greynoise_client._get_ip = MagicMock(side_effect=lambda ip: mock_greynoise_get(ip, greynoise_pool))
    greynoise_enrich = greynoise_client.enrich_ips(ip)

    abuseipdb_pool = json.loads(test_data("abuseipdb_1.1.1.1_red.json"))
    abuseipdb_client = AbuseIpDbClient("dummykey")
    abuseipdb_client._get_ip = MagicMock(side_effect=lambda ip: mock_abuseipdb_get(ip, abuseipdb_pool))
    abuseipdb_enrich = abuseipdb_client.enrich_ips(ip)

    urlhaus_pool = json.loads(test_data("urlhaus_1.1.1.1.json"))
    urlhaus_client = UrlHausClient()
    urlhaus_client._get_host = MagicMock(side_effect=lambda ip: mock_urlhaus_get(ip, urlhaus_pool))
    urlhaus_enrich = urlhaus_client.enrich_ips(ip)

    return IpAddressView(
        console=Console(),
        entity=IpAddress.model_validate(json.loads(test_data("vt_ip_1.1.1.1.json"))),
        geoasn=geoasn_enrich,
        whois=PTWhois.model_validate(json.loads(test_data("pt_whois_1.1.1.1.json"))),
        shodan=shodan_enrich,
        greynoise=greynoise_enrich,
        abuseipdb=abuseipdb_enrich,
        urlhaus=urlhaus_enrich,
    )


@pytest.fixture()
def view02(test_data, mock_ipwhois_get, mock_shodan_get_ip, mock_greynoise_get):
    """ 1.1.1.1 with Shodan and Greynoise. Test the whole IP panel. """
    ip = "1.1.1.1"
    geoasn_pool = json.loads(test_data("ipwhois_1.1.1.1.json"))
    geoasn_client = IpWhoisClient()
    geoasn_client._get_ipwhois = MagicMock(side_effect=lambda ip: mock_ipwhois_get(ip, geoasn_pool))
    geoasn_enrich = geoasn_client.enrich_ips(ip)

    shodan_pool = json.loads(test_data("shodan_1.1.1.1.json"))
    shodan_client = ShodanClient(MagicMock())
    shodan_client._get_ip = MagicMock(side_effect=lambda ip: mock_shodan_get_ip(ip, shodan_pool))
    shodan_enrich = shodan_client.enrich_ips(ip)

    greynoise_pool = json.loads(test_data("greynoise_1.1.1.1.json"))
    greynoise_client = GreynoiseClient("dummykey")
    greynoise_client._get_ip = MagicMock(side_effect=lambda ip: mock_greynoise_get(ip, greynoise_pool))
    greynoise_enrich = greynoise_client.enrich_ips(ip)

    return IpAddressView(
        console=Console(),
        entity=IpAddress.model_validate(json.loads(test_data("vt_ip_1.1.1.1.json"))),
        geoasn=geoasn_enrich,
        whois=MagicMock(),
        shodan=shodan_enrich,
        greynoise=greynoise_enrich,
        abuseipdb=MagicMock(),
        urlhaus=UrlHausMap.model_validate({}),
    )


@pytest.fixture()
def view03(test_data):
    """ 1.1.1.1 VT whois. Whois panel test only."""
    return IpAddressView(
        console=Console(),
        entity=MagicMock(),
        geoasn=MagicMock(),
        whois=VTWhois.model_validate(json.loads(test_data("vt_whois_1.1.1.1.json"))),
        shodan=MagicMock(),
        greynoise=MagicMock(),
        abuseipdb=MagicMock(),
        urlhaus=MagicMock(),
    )


@pytest.fixture()
def view04(test_data):
    """
    142.251.220.110. Test whole IP panel with 0 malicious, 0 reputation and no IP and
    Greynoise enrichment.
    """
    return IpAddressView(
        console=Console(),
        entity=IpAddress.model_validate(json.loads(test_data("vt_ip_142.251.220.110.json"))),
        geoasn=IpWhoisMap.model_validate({}),
        whois=MagicMock(),
        shodan=MagicMock(),
        greynoise=GreynoiseIpMap.model_validate({}),
        abuseipdb=MagicMock(),
        urlhaus=UrlHausMap.model_validate({}),
    )


@pytest.fixture()
def view05(test_data, mock_greynoise_get):
    """ 1.1.1.1 with alt Greynoise results. Test Greynoise only. """
    ip = "1.1.1.1"
    greynoise_pool = json.loads(test_data("greynoise_1.1.1.1_malicious.json"))
    greynoise_client = GreynoiseClient("dummykey")
    greynoise_client._get_ip = MagicMock(side_effect=lambda ip: mock_greynoise_get(ip, greynoise_pool))
    greynoise_enrich = greynoise_client.enrich_ips(ip)

    return IpAddressView(
        console=Console(),
        entity=IpAddress.model_validate(json.loads(test_data("vt_ip_1.1.1.1.json"))),
        geoasn=IpWhoisMap.model_validate({}),
        whois=MagicMock(),
        shodan=MagicMock(),
        greynoise=greynoise_enrich,
        abuseipdb=MagicMock(),
        urlhaus=UrlHausMap.model_validate({}),
    )


@pytest.fixture()
def view06(test_data, mock_greynoise_get):
    """ 1.1.1.1 with another alt Greynoise result (unknown class). Test Greynoise only. """
    ip = "1.1.1.1"
    greynoise_pool = json.loads(test_data("greynoise_1.1.1.1_unknown.json"))
    greynoise_client = GreynoiseClient("dummykey")
    greynoise_client._get_ip = MagicMock(side_effect=lambda ip: mock_greynoise_get(ip, greynoise_pool))
    greynoise_enrich = greynoise_client.enrich_ips(ip)

    return IpAddressView(
        console=Console(),
        entity=IpAddress.model_validate(json.loads(test_data("vt_ip_1.1.1.1.json"))),
        geoasn=IpWhoisMap.model_validate({}),
        whois=MagicMock(),
        shodan=MagicMock(),
        greynoise=greynoise_enrich,
        abuseipdb=MagicMock(),
        urlhaus=MagicMock(),
    )


@pytest.fixture()
def view07(test_data, mock_abuseipdb_get):
    """ 1.1.1.1 with green AbuseIPDB score. Test AbuseIPDB only. """
    ip = "1.1.1.1"
    abuseipdb_pool = json.loads(test_data("abuseipdb_1.1.1.1_green.json"))
    abuseipdb_client = AbuseIpDbClient("dummykey")
    abuseipdb_client._get_ip = MagicMock(side_effect=lambda ip: mock_abuseipdb_get(ip, abuseipdb_pool))
    abuseipdb_enrich = abuseipdb_client.enrich_ips(ip)

    return IpAddressView(
        console=Console(),
        entity=IpAddress.model_validate(json.loads(test_data("vt_ip_1.1.1.1.json"))),
        geoasn=IpWhoisMap.model_validate({}),
        whois=MagicMock(),
        shodan=MagicMock(),
        greynoise=MagicMock(),
        abuseipdb=abuseipdb_enrich,
        urlhaus=MagicMock(),
    )


@pytest.fixture()
def view08(test_data, mock_abuseipdb_get):
    """ 1.1.1.1 with yellow AbuseIPDB score. Test AbuseIPDB only. """
    ip = "1.1.1.1"
    abuseipdb_pool = json.loads(test_data("abuseipdb_1.1.1.1_yellow.json"))
    abuseipdb_client = AbuseIpDbClient("dummykey")
    abuseipdb_client._get_ip = MagicMock(side_effect=lambda ip: mock_abuseipdb_get(ip, abuseipdb_pool))
    abuseipdb_enrich = abuseipdb_client.enrich_ips(ip)

    return IpAddressView(
        console=Console(),
        entity=IpAddress.model_validate(json.loads(test_data("vt_ip_1.1.1.1.json"))),
        geoasn=IpWhoisMap.model_validate({}),
        whois=MagicMock(),
        shodan=MagicMock(),
        greynoise=MagicMock(),
        abuseipdb=abuseipdb_enrich,
        urlhaus=MagicMock(),
    )


class TestView01:
    def test_ip_panel(self, view01, theme, display_timestamp):
        ip = view01.ip_panel()
        assert type(ip) is Panel
        assert ip.title == Text("1.1.1.1")
        assert ip.title.style == theme.panel_title

        # Sections
        vt_section = ip.renderable.renderables[0]
        geoasn_section = ip.renderable.renderables[2]
        shodan_section = ip.renderable.renderables[4]
        urlhaus_section = ip.renderable.renderables[6]
        other_section = ip.renderable.renderables[8]

        # Line breaks between sections
        assert ip.renderable.renderables[1] == ""
        assert ip.renderable.renderables[3] == ""
        assert ip.renderable.renderables[5] == ""

        #
        # VT section
        #

        # Heading
        assert vt_section.renderables[0] == Text("VirusTotal")
        assert vt_section.renderables[0].style == theme.heading_h1

        # Table
        table = vt_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            Text(
                "Analysis:",
                spans=[Span(0, 8, 'link https://virustotal.com/gui/ip-address/1.1.1.1')],
            ),
            "Reputation:",
            "Updated:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                "4/94 malicious\nCMC Threat Intelligence, Comodo Valkyrie Verdict, CRDF, Blueliv",
                spans=[
                    Span(0,  14, theme.error),
                    Span(15, 38, theme.vendor_list),
                    Span(38, 40, "default"),
                    Span(40, 63, theme.vendor_list),
                    Span(63, 65, "default"),
                    Span(65, 69, theme.vendor_list),
                    Span(69, 71, "default"),
                    Span(71, 78, theme.vendor_list),
                ]
            ),
            Text("134"),
            display_timestamp("2022-09-03T06:47:04Z"),
        ]

        #
        # IP location and ASN section
        #

        # Heading
        assert geoasn_section.renderables[0] == Text("IPWhois")
        assert geoasn_section.renderables[0].style == theme.heading_h1

        # Table
        table = geoasn_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "ASN:",
            "ISP:",
            "Location:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                "13335 (APNIC and Cloudflare DNS Resolver project)",
                spans=[Span(7, 48, theme.asn_org)],
            ),
            "Cloudflare, Inc.",
            Text(
                "Sydney, New South Wales, Australia",
                spans=[
                    Span(6, 8, "default"),
                    Span(23, 25, "default"),
                ]
            ),
        ]

        #
        # Shodan section
        #

        # Heading
        assert shodan_section.renderables[0] == Text("Shodan")
        assert shodan_section.renderables[0].style == theme.heading_h1

        # Table
        table = shodan_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            Text(
                "Services:",
                spans=[Span(0, 8, 'link https://www.shodan.io/host/1.1.1.1')],
            ),
            "Last Scan:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                (
                    "Cisco router tftpd (69/udp)\nCloudFlare (80/tcp, 8080/tcp, 8880/tcp)\n"
                    "DrayTek Vigor Router (443/tcp)\nOther (53/tcp, 53/udp, 161/udp, "
                    "2082/tcp, 2083/tcp, 2086/tcp, 2087/tcp, 8443/tcp)"
                ),
                spans=[
                    Span(0, 18, theme.product),
                    Span(20, 22, theme.port),
                    Span(22, 26, theme.transport),
                    Span(28, 38, theme.product),
                    Span(40, 42, theme.port),
                    Span(42, 46, theme.transport),
                    Span(46, 48, "default"),
                    Span(48, 52, theme.port),
                    Span(52, 56, theme.transport),
                    Span(56, 58, "default"),
                    Span(58, 62, theme.port),
                    Span(62, 66, theme.transport),
                    Span(68, 88, theme.product),
                    Span(90, 93, theme.port),
                    Span(93, 97, theme.transport),
                    Span(99, 104, theme.product),
                    Span(106, 108, theme.port),
                    Span(108, 112, theme.transport),
                    Span(112, 114, "default"),
                    Span(114, 116, theme.port),
                    Span(116, 120, theme.transport),
                    Span(120, 122, "default"),
                    Span(122, 125, theme.port),
                    Span(125, 129, theme.transport),
                    Span(129, 131, "default"),
                    Span(131, 135, theme.port),
                    Span(135, 139, theme.transport),
                    Span(139, 141, "default"),
                    Span(141, 145, theme.port),
                    Span(145, 149, theme.transport),
                    Span(149, 151, "default"),
                    Span(151, 155, theme.port),
                    Span(155, 159, theme.transport),
                    Span(159, 161, "default"),
                    Span(161, 165, theme.port),
                    Span(165, 169, theme.transport),
                    Span(169, 171, "default"),
                    Span(171, 175, theme.port),
                    Span(175, 179, theme.transport),
                ]
            ),
            display_timestamp("2022-09-04T01:03:56Z"),
        ]

        #
        # Urlhaus section
        #

        # Heading
        assert urlhaus_section.renderables[0] == Text("URLhaus")
        assert urlhaus_section.renderables[0].style == theme.heading_h1

        # Table
        table = urlhaus_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            Text(
                "Malware URLs:",
                spans=[Span(0, 12, 'link https://urlhaus.abuse.ch/host/1.1.1.1/')]
            ),
            "Blocklists:",
            "Tags:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        print(table.columns[1])
        assert table.columns[1]._cells == [
            Text(
                "10 online (10 total)",
                spans=[
                    Span(0, 9, theme.error),
                    Span(9, 20, theme.table_value),
                ]
            ),
            Text(
                "not listed in spamhaus\nnot listed in surbl",
                spans=[
                    Span(0, 10, theme.urlhaus_bl_low),
                    Span(14, 22, theme.urlhaus_bl_name),
                    Span(23, 33, theme.urlhaus_bl_low),
                    Span(37, 42, theme.urlhaus_bl_name),
                ]
            ),
            Text(
                "elf, mirai",
                spans=[
                    Span(0, 3, theme.tags),
                    Span(3, 5, "default"),
                    Span(5, 10, theme.tags),
                ]
            ),
        ]

        #
        # Other section
        #

        # Heading
        assert other_section.renderables[0] == Text("Other")
        assert other_section.renderables[0].style == theme.heading_h1

        # Table
        table = other_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            Text(
                "GreyNoise:",
                spans=[Span(0, 9, "link https://viz.greynoise.io/riot/1.1.1.1")]
            ),
            Text(
                "AbuseIPDB:",
                spans=[Span(0, 9, "link https://www.abuseipdb.com/check/1.1.1.1")]
            )
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                "✓ riot  ✗ noise  ✓ benign",
                spans=[
                    Span(0, 1, theme.info),
                    Span(2, 6, theme.tags),
                    Span(8, 9, theme.warn),
                    Span(10, 15, theme.tags),
                    Span(17, 18, theme.info),
                    Span(19, 25, theme.tags_green),
                ]
            ),
            Text(
                "100 confidence score (567 reports)",
                spans=[
                    Span(0, 3, theme.error),
                    Span(3, 20, "red"),
                    Span(20, 34, theme.table_value),
                ]
            ),
        ]

    def test_whois_panel(self, view01, theme):
        whois = view01.whois_panel()
        assert type(whois) is Panel
        assert whois.title is None

        # Sections
        title = whois.renderable.renderables[0]
        content = whois.renderable.renderables[1]

        # Title
        assert title == Text("Whois")
        assert title.style == theme.heading_h1

        # Heading
        assert content.renderables[0] == Text(
            "1.1.1.0",
            spans=[Span(0, 7, f"{theme.heading_h2} link https://community.riskiq.com/search/1.1.1.0/whois")]
        )

        # Table
        table = content.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "Registrar:",
            "Organization:",
            "Name:",
            "Email:",
            "Phone:",
            "Street:",
            "Registered:",
            "Updated:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert [str(c) for c in table.columns[1]._cells] == [
            "APNIC",
            "APNIC Research and Development",
            "APNIC Research and Development",
            "helpdesk@apnic.net",
            "+61-7-38583100",
            "6 cordelia st",
            "2011-08-10T23:12:35Z",
            "2020-07-15T13:10:57Z",
        ]

    def test_print(self, view01):
        view01.ip_panel = MagicMock()
        view01.whois_panel = MagicMock()
        view01.console.print = MagicMock()

        view01.print()
        view01.ip_panel.assert_called_once()
        view01.whois_panel.assert_called_once()
        view01.console.print.assert_called_once()


class TestView02:
    def test_ip_panel(self, view02, theme, display_timestamp):
        ip = view02.ip_panel()
        assert type(ip) is Panel
        assert ip.title == Text("1.1.1.1")

        # Sections
        vt_section = ip.renderable.renderables[0]
        geoasn_section = ip.renderable.renderables[2]
        shodan_section = ip.renderable.renderables[4]
        other_section = ip.renderable.renderables[6]

        # Line breaks between sections
        assert ip.renderable.renderables[1] == ""
        assert ip.renderable.renderables[3] == ""

        #
        # VT section
        #

        # Heading
        assert vt_section.renderables[0] == Text("VirusTotal")
        assert vt_section.renderables[0].style == theme.heading_h1

        # Table
        table = vt_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            Text(
                "Analysis:",
                spans=[Span(0, 8, 'link https://virustotal.com/gui/ip-address/1.1.1.1')],
            ),
            "Reputation:",
            "Updated:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                "4/94 malicious\nCMC Threat Intelligence, Comodo Valkyrie Verdict, CRDF, Blueliv",
                spans=[
                    Span(0, 14, theme.error),
                    Span(15, 38, theme.vendor_list),
                    Span(38, 40, "default"),
                    Span(40, 63, theme.vendor_list),
                    Span(63, 65, "default"),
                    Span(65, 69, theme.vendor_list),
                    Span(69, 71, "default"),
                    Span(71, 78, theme.vendor_list),
                ]
            ),
            Text("134"),
            display_timestamp("2022-09-03T06:47:04Z"),
        ]

        #
        # IP location and ASN section
        #

        # Heading
        assert geoasn_section.renderables[0] == Text("IPWhois")
        assert geoasn_section.renderables[0].style == theme.heading_h1

        # Table
        table = geoasn_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "ASN:",
            "ISP:",
            "Location:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                "13335 (APNIC and Cloudflare DNS Resolver project)",
                spans=[Span(7, 48, theme.asn_org)],
            ),
            "Cloudflare, Inc.",
            Text(
                "Sydney, New South Wales, Australia",
                spans=[
                    Span(6, 8, "default"),
                    Span(23, 25, "default")
                ],
            ),
        ]

        #
        # Shodan section
        #

        # Heading
        assert shodan_section.renderables[0] == Text("Shodan")
        assert shodan_section.renderables[0].style == theme.heading_h1

        # Table
        table = shodan_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            Text(
                "Services:",
                spans=[Span(0, 8, 'link https://www.shodan.io/host/1.1.1.1')],
            ),
            "Last Scan:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                (
                    "Cisco router tftpd (69/udp)\nCloudFlare (80/tcp, 8080/tcp, 8880/tcp)\n"
                    "DrayTek Vigor Router (443/tcp)\nOther (53/tcp, 53/udp, 161/udp, "
                    "2082/tcp, 2083/tcp, 2086/tcp, 2087/tcp, 8443/tcp)"
                ),
                spans=[
                    Span(0, 18, theme.product),
                    Span(20, 22, theme.port),
                    Span(22, 26, theme.transport),
                    Span(28, 38, theme.product),
                    Span(40, 42, theme.port),
                    Span(42, 46, theme.transport),
                    Span(46, 48, "default"),
                    Span(48, 52, theme.port),
                    Span(52, 56, theme.transport),
                    Span(56, 58, "default"),
                    Span(58, 62, theme.port),
                    Span(62, 66, theme.transport),
                    Span(68, 88, theme.product),
                    Span(90, 93, theme.port),
                    Span(93, 97, theme.transport),
                    Span(99, 104, theme.product),
                    Span(106, 108, theme.port),
                    Span(108, 112, theme.transport),
                    Span(112, 114, "default"),
                    Span(114, 116, theme.port),
                    Span(116, 120, theme.transport),
                    Span(120, 122, "default"),
                    Span(122, 125, theme.port),
                    Span(125, 129, theme.transport),
                    Span(129, 131, "default"),
                    Span(131, 135, theme.port),
                    Span(135, 139, theme.transport),
                    Span(139, 141, "default"),
                    Span(141, 145, theme.port),
                    Span(145, 149, theme.transport),
                    Span(149, 151, "default"),
                    Span(151, 155, theme.port),
                    Span(155, 159, theme.transport),
                    Span(159, 161, "default"),
                    Span(161, 165, theme.port),
                    Span(165, 169, theme.transport),
                    Span(169, 171, "default"),
                    Span(171, 175, theme.port),
                    Span(175, 179, theme.transport),
                ]
            ),
            display_timestamp("2022-09-04T01:03:56Z"),
        ]

        #
        # Other section
        #

        # Heading
        assert other_section.renderables[0] == Text("Other")
        assert other_section.renderables[0].style == theme.heading_h1

        # Table
        table = other_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            Text(
                "GreyNoise:",
                spans=[Span(0, 9, "link https://viz.greynoise.io/riot/1.1.1.1")]
            ),
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                "✓ riot  ✗ noise  ✓ benign",
                spans=[
                    Span(0, 1, theme.info),
                    Span(2, 6, theme.tags),
                    Span(8, 9, theme.warn),
                    Span(10, 15, theme.tags),
                    Span(17, 18, theme.info),
                    Span(19, 25, theme.tags_green),
                ]
            ),
        ]


class TestView03:
    def test_whois_panel(self, view03, theme):
        whois = view03.whois_panel()
        assert type(whois) is Panel
        assert whois.title is None

        # Sections
        title = whois.renderable.renderables[0]
        content = whois.renderable.renderables[1]

        # Title
        assert title == Text("Whois")
        assert title.style == theme.heading_h1

        # Content
        assert content.renderables[0] == Text(
            "one.one",
            spans=[Span(0, 7, 'bold yellow')],
        )

        table = content.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "Registrar:",
            "Organization:",
            "Name:",
            "Email:",
            "Phone:",
            "Street:",
            "City:",
            "Country:",
            "Postcode:",
            "Nameservers:",
            "DNSSEC:",
            "Registered:",
            "Updated:",
            "Expires:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert [str(c) for c in table.columns[1]._cells] == [
            "One.com A/S - ONE",
            "One.com A/S",
            "REDACTED FOR PRIVACY",
            (
                "Please query the RDDS service of the Registrar of Record identified in this "
                "output for information on how to contact the Registrant, Admin, or Tech "
                "contact of the queried domain name."
            ),
            "REDACTED FOR PRIVACY",
            "REDACTED FOR PRIVACY",
            "REDACTED FOR PRIVACY",
            "dk",
            "REDACTED FOR PRIVACY",
            (
                "* a response from the service that a domain name is 'available', does not "
                "guarantee that is able to be registered,, * we may restrict, suspend or "
                "terminate your access to the service at any time, and, * the copying, "
                "compilation, repackaging, dissemination or other use of the information "
                "provided by the service is not permitted, without our express written "
                "consent., this information has been prepared and published in order to "
                "represent administrative and technical management of the tld., we may "
                "discontinue or amend any part or the whole of these terms of service from "
                "time to time at our absolute discretion."
            ),
            "signedDelegation",
            "2015-05-20T12:15:44Z",
            "2021-07-04T12:15:49Z",
            "2022-05-20T12:15:44Z",
        ]


class TestView04:
    def test_ip_panel(self, view04, theme, display_timestamp):
        ip = view04.ip_panel()
        assert type(ip) is Panel
        assert ip.title == Text("142.251.220.110")

        # Sections
        vt_section = ip.renderable.renderables[0]

        #
        # VT section
        #

        # Heading
        assert vt_section.renderables[0] == Text("VirusTotal")

        # Table
        table = vt_section.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == theme.table_field
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            Text(
                "Analysis:",
                spans=[Span(0, 8, 'link https://virustotal.com/gui/ip-address/142.251.220.110')],
            ),
            "Reputation:",
            "Updated:",
        ]
        assert table.columns[1].style == theme.table_value
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text("0/93 malicious", spans=[Span(0, 14, theme.info)]),
            Text("0"),
            display_timestamp("2022-09-03T16:58:45Z"),
        ]


class TestView05:
    def test_ip_panel_greynoise_only(self, view05, theme):
        ip = view05.ip_panel()

        other_section = ip.renderable.renderables[2]

        # Table
        table = other_section.renderables[1]
        assert table.columns[1]._cells[-1] == Text(
            "✗ riot  ✓ noise  ! malicious",
            spans=[
                Span(0, 1, theme.warn),
                Span(2, 6, theme.tags),
                Span(8, 9, theme.info),
                Span(10, 15, theme.tags),
                Span(17, 18, theme.error),
                Span(19, 28, theme.tags_red),
            ]
        )


class TestView06:
    def test_ip_panel_greynoise_only(self, view06, theme):
        ip = view06.ip_panel()

        other_section = ip.renderable.renderables[2]

        # Table
        table = other_section.renderables[1]
        assert table.columns[1]._cells[-1] == Text(
            "✓ riot  ✗ noise  ? unknown",
            spans=[
                Span(0, 1, theme.info),
                Span(2, 6, theme.tags),
                Span(8, 9, theme.warn),
                Span(10, 15, theme.tags),
                Span(19, 26, theme.tags),
            ]
        )


class TestAbuseIpDbOnly:
    def test_abuseipdb_green(self, view07, theme):
        ip = view07.ip_panel()

        other_section = ip.renderable.renderables[2]

        # Table
        table = other_section.renderables[1]
        assert table.columns[1]._cells[-1] == Text(
            "0 confidence score",
            spans=[
                Span(0, 1, theme.info),
                Span(1, 18, "green"),
            ]
        )

    def test_abuseipdb_yellow(self, view08, theme):
        ip = view08.ip_panel()

        other_section = ip.renderable.renderables[2]

        # Table
        table = other_section.renderables[1]
        assert table.columns[1]._cells[-1] == Text(
            "30 confidence score (567 reports)",
            spans=[
                Span(0, 2, theme.warn),
                Span(2, 19, "yellow"),
                Span(19, 33, theme.table_value),
            ]
        )



"""
Type aliases
"""
from typing import Union

from wtfis.models.abuseipdb import AbuseIpDbMap
from wtfis.models.greynoise import GreynoiseIpMap
from wtfis.models.ipwhois import IpWhois, IpWhoisMap
from wtfis.models.shodan import ShodanIpMap
from wtfis.models.urlhaus import UrlHausMap


# IP enrichment map types
IpEnrichmentType = Union[
    AbuseIpDbMap,
    GreynoiseIpMap,
    IpWhoisMap,
    ShodanIpMap,
    UrlHausMap,
]

# Domain/FQDN enrichment map types
DomainEnrichmentType = Union[
    UrlHausMap,
]

# IP geolocation and ASN types
IpGeoAsnType = Union[
    IpWhois,
]

IpGeoAsnMapType = Union[
    IpWhoisMap,
]



import json
import pytest
from unittest.mock import MagicMock, patch

from shodan import APIError

from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.base import requests
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.ip2whois import Ip2WhoisClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.passivetotal import PTClient
from wtfis.clients.shodan import Shodan, ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient
from wtfis.models.ipwhois import IpWhoisMap


@pytest.fixture()
def abuseipdb_client():
    return AbuseIpDbClient("dummykey")


@pytest.fixture()
def greynoise_client():
    return GreynoiseClient("dummykey")


@pytest.fixture()
def ip2whois_client():
    return Ip2WhoisClient("dummykey")


@pytest.fixture()
def ipwhois_client():
    return IpWhoisClient()


@pytest.fixture()
def passivetotal_client():
    return PTClient("dummyuser", "dummykey")


@pytest.fixture()
def shodan_client():
    return ShodanClient("dummykey")


@pytest.fixture()
def urlhaus_client():
    return UrlHausClient()


@pytest.fixture()
def virustotal_client():
    return VTClient("dummykey")


class TestAbuseIpDbClient:
    def test_init(self, abuseipdb_client):
        assert abuseipdb_client.name == "AbuseIPDB"
        assert abuseipdb_client.api_key == "dummykey"

    @patch.object(requests.Session, "get")
    def test_enrich_ips(self, mock_requests_get, test_data, abuseipdb_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("abuseipdb_1.1.1.1_raw.json"))
        mock_requests_get.return_value = mock_resp

        abuseipdb = abuseipdb_client.enrich_ips("thisdoesntmatter").root["1.1.1.1"]

        assert abuseipdb.ip_address == "1.1.1.1"
        assert abuseipdb.abuse_confidence_score == 100
        assert abuseipdb.usage_type == "Data Center/Web Hosting/Transit"
        assert abuseipdb.total_reports == 567
        assert abuseipdb.num_distinct_users == 123


class TestIp2WhoisClient:
    def test_init(self, ip2whois_client):
        assert ip2whois_client.api_key == "dummykey"
        assert ip2whois_client.name == "IP2Whois"

    @patch.object(requests.Session, "get")
    def test_get_whois(self, mock_requests_get, test_data, ip2whois_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("ip2whois_whois_hotmail.json"))
        mock_requests_get.return_value = mock_resp

        whois = ip2whois_client.get_whois("thisdoesntmatter")

        assert whois.domain == "hotmail.com"

    @patch.object(requests.Session, "get")
    def test_get_whois_http_error(self, mock_requests_get, ip2whois_client):
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            # 404 - no exception
            mock_resp.status_code = 404
            mock_resp_json.return_value = {"error": {"error_code": 10006}}
            mock_requests_get.return_value = mock_resp
            assert ip2whois_client.get_whois("thisdoesntmatter").source == "ip2whois"

            # 400 - no exception
            mock_resp.status_code = 400
            mock_resp_json.return_value = {"error": {"error_code": 10007}}
            mock_requests_get.return_value = mock_resp
            assert ip2whois_client.get_whois("thisdoesntmatter").source == "ip2whois"

            # 400 - with exception
            mock_resp.status_code = 400
            mock_resp_json.return_value = {"error": {"error_code": 10008}}
            mock_requests_get.return_value = mock_resp
            with pytest.raises(requests.exceptions.HTTPError) as err:
                ip2whois_client.get_whois("thisdoesntmatter")
            assert err.value.response.json()["error"]["error_code"] == 10008


class TestGreynoiseClient:
    def test_init(self, greynoise_client):
        assert greynoise_client.name == "Greynoise"
        assert greynoise_client.api_key == "dummykey"


class TestIpWhoisClient:
    def test_init(self, ipwhois_client):
        assert ipwhois_client.name == "IPWhois"

    @patch.object(requests.Session, "get")
    def test_get_ipwhois(self, mock_requests_get, test_data, ipwhois_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("ipwhois_raw_10.0.0.1.json"))
        mock_requests_get.return_value = mock_resp

        whois = ipwhois_client.enrich_ips("thisdoesntmatter")

        assert whois == IpWhoisMap.model_validate({})


class TestPassivetotalClient:
    def test_init(self, passivetotal_client):
        assert passivetotal_client.s.auth == ("dummyuser", "dummykey")
        assert passivetotal_client.name == "Passivetotal"

    @patch.object(requests.Session, "get")
    def test_get_passive_dns(self, mock_requests_get, test_data, passivetotal_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("pt_passive_dns_gist.json"))
        mock_requests_get.return_value = mock_resp

        dns = passivetotal_client.get_passive_dns("thisdoesntmatter")

        assert dns["queryValue"] == "gist.github.com"
        assert dns["queryType"] == "domain"
        assert dns["totalRecords"] == 30


class TestShodanClient:
    def test_init(self, shodan_client):
        assert shodan_client.s.api_key == "dummykey"
        assert shodan_client.name == "Shodan"

    @patch.object(Shodan, "host")
    def test_get_ip_apierror_invalid_key(self, mock_shodan_host, shodan_client):
        """ Test invalid API key APIError """
        mock_shodan_host.side_effect = APIError("Invalid API key")

        with pytest.raises(APIError) as e:
            shodan_client.enrich_ips("thisdoesntmatter")

        assert e.type == APIError
        assert str(e.value) == "Invalid API key"

    @patch.object(Shodan, "host")
    def test_get_ip_apierror_other(self, mock_shodan_host, shodan_client):
        """ Test other invalid API key APIError """
        mock_shodan_host.side_effect = APIError("Some other error")

        with pytest.raises(APIError) as e:
            shodan_client.enrich_ips("thisdoesntmatter")

        assert e.type == APIError
        assert str(e.value) == "Some other error"


class TestUrlhausClient:
    def test_init(self, urlhaus_client):
        assert urlhaus_client.name == "URLhaus"

    @patch.object(requests.Session, "post")
    def test_enrich_ips(self, mock_requests_post, test_data, urlhaus_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("urlhaus_1.1.1.1.json"))["1.1.1.1"]
        mock_requests_post.return_value = mock_resp

        urlhaus = urlhaus_client.enrich_ips("thisdoesntmatter").root["1.1.1.1"]

        assert urlhaus.host == "1.1.1.1"
        assert urlhaus.url_count == 10
        assert urlhaus.urlhaus_reference == "https://urlhaus.abuse.ch/host/1.1.1.1/"


class TestVirustotalClient:
    def test_init(self, virustotal_client):
        assert virustotal_client.s.headers["x-apikey"] == "dummykey"
        assert virustotal_client.s.headers["Accept"] == "application/json"
        assert virustotal_client.name == "Virustotal"

    @patch.object(requests.Session, "get")
    def test_get_whois_domain(self, mock_requests_get, test_data, virustotal_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("vt_whois_gist.json"))
        mock_requests_get.return_value = mock_resp

        whois = virustotal_client.get_whois("gist.github.com")

        assert whois.domain == "github.com"
        assert whois.registrar == "MarkMonitor Inc."

    @patch.object(requests.Session, "get")
    def test_get_whois_ip(self, mock_requests_get, test_data, virustotal_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("vt_whois_1.1.1.1.json"))
        mock_requests_get.return_value = mock_resp

        whois = virustotal_client.get_whois("1.1.1.1")

        assert whois.domain == "one.one"
        assert whois.registrar == "One.com A/S - ONE"



from pydantic import Field, field_validator, model_validator
from typing import List, Optional

from wtfis.models.base import WhoisBase


class Whois(WhoisBase):
    source: str = "ip2whois"
    domain: str = ""
    registrar: Optional[str] = None
    organization: Optional[str] = None
    name: Optional[str] = None
    email: Optional[str] = None
    phone: Optional[str] = None
    street: Optional[str] = Field(None, alias="street_address")
    city: Optional[str] = None
    state: Optional[str] = Field(None, alias="region")
    country: Optional[str] = None
    postal_code: Optional[str] = Field(None, alias="zip_code")
    name_servers: List[str] = Field([], alias="nameservers")
    date_created: Optional[str] = Field(None, alias="create_date")
    date_changed: Optional[str] = Field(None, alias="update_date")
    date_expires: Optional[str] = Field(None, alias="expire_date")
    whois_server: Optional[str] = None
    dnssec: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def extract_registrant(cls, v):
        """ Surface registrant fields to root level """
        registrant = v.pop("registrant", {})
        if not registrant:
            return v
        for field in [
            "organization", "name", "email", "phone",
            "street_address", "city", "region", "country", "zip_code",
        ]:
            v[field] = registrant.get(field)
        return v

    @field_validator("registrar", mode="before")
    @classmethod
    def transform_registrar(cls, v):
        """ Convert registrar from dict to simply registrar.name """
        return v.get("name") if v else v



from shodan import Shodan
from typing import Optional

from wtfis.clients.base import BaseClient, BaseIpEnricherClient
from wtfis.models.shodan import ShodanIp, ShodanIpMap


class ShodanClient(BaseClient, BaseIpEnricherClient):
    """
    Shodan client
    """
    def __init__(self, api_key: str) -> None:
        self.s = Shodan(api_key)

    @property
    def name(self) -> str:
        return "Shodan"

    def _get_ip(self, ip: str) -> Optional[ShodanIp]:
        return ShodanIp.model_validate(self.s.host(ip, minify=False))

    def enrich_ips(self, *ips: str) -> ShodanIpMap:
        shodan_map = {}
        for ip in ips:
            ip_data = self._get_ip(ip)
            if ip_data:
                shodan_map[ip_data.ip_str] = ip_data
        return ShodanIpMap.model_validate(shodan_map)



import json
import os
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from dotenv import load_dotenv
from rich.console import Console
from rich.progress import (BarColumn, Progress, SpinnerColumn,
                           TaskProgressColumn, TextColumn, TimeElapsedColumn)

from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.ip2whois import Ip2WhoisClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.passivetotal import PTClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient
from wtfis.handlers.domain import DomainHandler
from wtfis.handlers.ip import IpAddressHandler
from wtfis.main import (generate_entity_handler, generate_view, main,
                        parse_args, parse_env)
from wtfis.models.virustotal import Domain, IpAddress
from wtfis.ui.view import DomainView, IpAddressView

POSSIBLE_ENV_VARS = [
    "VT_API_KEY",
    "PT_API_KEY",
    "PT_API_USER",
    "IP2WHOIS_API_KEY",
    "SHODAN_API_KEY",
    "GREYNOISE_API_KEY",
    "ABUSEIPDB_API_KEY",
    "WTFIS_DEFAULTS",
]


def unset_env_vars():
    for var in POSSIBLE_ENV_VARS:
        try:
            del os.environ[var]
        except KeyError:
            pass


def fake_load_dotenv(tmp_path, fake_env_vars):
    content = [f"{k} = {v}" for k, v in fake_env_vars.items()]
    path = tmp_path / ".env.wtfis"
    path.write_text("\n".join(content))

    def fake(*_):
        load_dotenv(path)

    return fake


def simulate_progress(console):
    return Progress(
        SpinnerColumn(finished_text="[green]✓"),
        TextColumn("[bold]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        TimeElapsedColumn(),
        console=console,
    )


@pytest.fixture()
def fake_load_dotenv_1(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "PT_API_KEY": "bar",
        "PT_API_USER": "baz@example.com",
        "IP2WHOIS_API_KEY": "alice",
        "SHODAN_API_KEY": "hunter2",
        "GREYNOISE_API_KEY": "upupdowndown",
        "ABUSEIPDB_API_KEY": "dummy",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_2(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "SHODAN_API_KEY": "hunter2",
        "WTFIS_DEFAULTS": "-s -1",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_3(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "WTFIS_DEFAULTS": "--no-color",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_4(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "GREYNOISE_API_KEY": "bar",
        "ABUSEIPDB_API_KEY": "dummy",
        "WTFIS_DEFAULTS": "-g -u -a",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_vt_whois(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_ip2whois(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "IP2WHOIS_API_KEY": "alice",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


class TestArgs:
    def test_basic(self):
        with patch("sys.argv", [
            "main",
            "www.example.com",
        ]):
            args = parse_args()
            assert args.entity == "www.example.com"
            assert args.max_resolutions == 3
            assert args.no_color is False
            assert args.one_column is False
            assert args.use_shodan is False
            assert args.use_greynoise is False
            assert args.use_urlhaus is False

    def test_display(self):
        with patch("sys.argv", [
            "main",
            "www.example.com",
            "-n",
            "-1",
        ]):
            args = parse_args()
            assert args.no_color is True
            assert args.one_column is True

    def test_max_resolutions_ok(self):
        with patch("sys.argv", [
            "main",
            "www.example.com",
            "-m",
            "8",
        ]):
            args = parse_args()
            assert args.max_resolutions == 8

    def test_max_resolutions_error_1(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch("sys.argv", [
                "main",
                "www.example.com",
                "-m",
                "11",
            ]):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err == "usage: main [-h]\nmain: error: Maximum --max-resolutions value is 10\n"
        assert e.type == SystemExit
        assert e.value.code == 2

    def test_max_resolutions_error_2(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch("sys.argv", [
                "main",
                "1.1.1.1",
                "-m",
                "5",
            ]):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err == "usage: main [-h]\nmain: error: --max-resolutions is not applicable to IPs\n"
        assert e.type == SystemExit
        assert e.value.code == 2

    def test_shodan_ok(self):
        os.environ["SHODAN_API_KEY"] = "foo"
        with patch("sys.argv", [
            "main",
            "www.example.com",
            "-s",
        ]):
            args = parse_args()
            assert args.use_shodan is True
        del os.environ["SHODAN_API_KEY"]

    def test_shodan_error(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch("sys.argv", [
                "main",
                "www.example.com",
                "-s",
            ]):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err == "usage: main [-h]\nmain: error: SHODAN_API_KEY is not set\n"
        assert e.type == SystemExit
        assert e.value.code == 2

    def test_greynoise_ok(self):
        os.environ["GREYNOISE_API_KEY"] = "foo"
        with patch("sys.argv", [
            "main",
            "www.example.com",
            "-g",
        ]):
            args = parse_args()
            assert args.use_greynoise is True
        del os.environ["GREYNOISE_API_KEY"]

    def test_greynoise_error(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch("sys.argv", [
                "main",
                "www.example.com",
                "-g",
            ]):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err == "usage: main [-h]\nmain: error: GREYNOISE_API_KEY is not set\n"
        assert e.type == SystemExit
        assert e.value.code == 2

    def test_urlhaus(self):
        with patch("sys.argv", [
            "main",
            "www.example.com",
            "-u",
        ]):
            args = parse_args()
            assert args.use_urlhaus is True

    def test_abuseipdb_ok(self):
        os.environ["ABUSEIPDB_API_KEY"] = "foo"
        with patch("sys.argv", [
            "main",
            "1.1.1.1",
            "-a",
        ]):
            args = parse_args()
            assert args.use_abuseipdb is True
        del os.environ["ABUSEIPDB_API_KEY"]

    def test_abuseipdb_error(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch("sys.argv", [
                "main",
                "1.1.1.1",
                "-a",
            ]):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err == "usage: main [-h]\nmain: error: ABUSEIPDB_API_KEY is not set\n"
        assert e.type == SystemExit
        assert e.value.code == 2


class TestEnvs:
    def test_env_file(self, fake_load_dotenv_1):
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_1):
            parse_env()
            assert os.environ["VT_API_KEY"] == "foo"
            assert os.environ["PT_API_KEY"] == "bar"
            assert os.environ["PT_API_USER"] == "baz@example.com"
            assert os.environ["IP2WHOIS_API_KEY"] == "alice"
            assert os.environ["SHODAN_API_KEY"] == "hunter2"
            assert os.environ["GREYNOISE_API_KEY"] == "upupdowndown"
            assert os.environ["ABUSEIPDB_API_KEY"] == "dummy"
        unset_env_vars()

    @patch("wtfis.main.load_dotenv", MagicMock())
    def test_required_env_vars(self):
        os.environ["VT_API_KEY"] = "foo"
        parse_env()
        unset_env_vars()

    @patch("wtfis.main.load_dotenv", MagicMock())
    @patch("wtfis.main.Path.exists")
    def test_error(self, mock_exists, capsys):
        mock_exists.return_value = False

        with pytest.raises(SystemExit) as e:
            parse_env()

        capture = capsys.readouterr()

        assert capture.err == (
            "Error: Environment variable VT_API_KEY not set\n"
            f"Env file {Path().home() / '.env.wtfis'} was not found either. Did you forget?\n"
        )
        assert e.type == SystemExit
        assert e.value.code == 1


class TestDefaults:
    def test_defaults_1(self, fake_load_dotenv_2):
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_2):
            with patch("sys.argv", [
                "main",
                "www.example.com",
            ]):
                parse_env()
                args = parse_args()
                assert args.entity == "www.example.com"
                assert args.max_resolutions == 3
                assert args.no_color is False
                assert args.one_column is True
                assert args.use_shodan is True
                assert args.use_greynoise is False
                assert args.use_urlhaus is False
        unset_env_vars()

    def test_defaults_2(self, fake_load_dotenv_2):
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_2):
            with patch("sys.argv", [
                "main",
                "www.example.com",
                "-s",
            ]):
                parse_env()
                args = parse_args()
                assert args.entity == "www.example.com"
                assert args.max_resolutions == 3
                assert args.no_color is False
                assert args.one_column is True
                assert args.use_shodan is False
                assert args.use_greynoise is False
                assert args.use_urlhaus is False
        unset_env_vars()

    def test_defaults_3(self, fake_load_dotenv_3):
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_3):
            with patch("sys.argv", [
                "main",
                "www.example.com",
            ]):
                parse_env()
                args = parse_args()
                assert args.entity == "www.example.com"
                assert args.max_resolutions == 3
                assert args.no_color is True
                assert args.one_column is False
                assert args.use_shodan is False
                assert args.use_greynoise is False
                assert args.use_urlhaus is False
        unset_env_vars()

    def test_defaults_4(self, fake_load_dotenv_4):
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_4):
            with patch("sys.argv", [
                "main",
                "1.1.1.1",
                "-u",
            ]):
                parse_env()
                args = parse_args()
                assert args.entity == "1.1.1.1"
                assert args.no_color is False
                assert args.one_column is False
                assert args.use_shodan is False
                assert args.use_greynoise is True
                assert args.use_urlhaus is False
                assert args.use_abuseipdb is True
        unset_env_vars()


class TestGenEntityHandler:
    """ Tests for the generate_entity_handler function """
    @patch("sys.argv", ["main", "www.example[.]com"])
    def test_handler_domain_1(self, fake_load_dotenv_1):
        """ Domain with default params """
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_1):
            parse_env()
            console = Console()
            progress = simulate_progress(console),
            entity = generate_entity_handler(parse_args(), console, progress)
        assert isinstance(entity, DomainHandler)
        assert entity.entity == "www.example.com"
        assert entity.max_resolutions == 3
        assert entity.console == console
        assert entity.progress == progress
        assert isinstance(entity._vt, VTClient)
        assert isinstance(entity._geoasn, IpWhoisClient)
        assert isinstance(entity._whois, PTClient)
        assert entity._shodan is None
        assert entity._greynoise is None
        assert entity._urlhaus is None
        assert entity._abuseipdb is None
        unset_env_vars()

    @patch("sys.argv", ["main", "www.example[.]com", "-s", "-g", "-u", "-m", "5"])
    def test_handler_domain_2(self, fake_load_dotenv_1):
        """ Domain with Shodan and Greynoise and non-default max_resolutions """
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_1):
            parse_env()
            console = Console()
            progress = simulate_progress(console),
            entity = generate_entity_handler(parse_args(), console, progress)
        assert entity.max_resolutions == 5
        assert isinstance(entity._geoasn, IpWhoisClient)
        assert isinstance(entity._whois, PTClient)
        assert isinstance(entity._shodan, ShodanClient)
        assert isinstance(entity._greynoise, GreynoiseClient)
        assert isinstance(entity._urlhaus, UrlHausClient)
        unset_env_vars()

    @patch("sys.argv", ["main", "www.example[.]com"])
    def test_handler_domain_3(self, fake_load_dotenv_vt_whois):
        """ Domain using default Ip2Whois for whois """
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_vt_whois):
            parse_env()
            console = Console()
            progress = simulate_progress(console),
            entity = generate_entity_handler(parse_args(), console, progress)
        assert isinstance(entity._whois, VTClient)
        unset_env_vars()

    @patch("sys.argv", ["main", "www.example[.]com"])
    def test_handler_domain_4(self, fake_load_dotenv_ip2whois):
        """ Domain using default Ip2Whois for whois """
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_ip2whois):
            parse_env()
            console = Console()
            progress = simulate_progress(console),
            entity = generate_entity_handler(parse_args(), console, progress)
        assert isinstance(entity._whois, Ip2WhoisClient)
        unset_env_vars()

    @patch("sys.argv", ["main", "1[.]1[.]1[.]1"])
    def test_handler_ip_1(self, fake_load_dotenv_1):
        """ IP with default params """
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_1):
            parse_env()
            console = Console()
            progress = simulate_progress(console),
            entity = generate_entity_handler(parse_args(), console, progress)
        assert isinstance(entity, IpAddressHandler)
        assert entity.entity == "1.1.1.1"
        assert entity.console == console
        assert entity.progress == progress
        assert isinstance(entity._vt, VTClient)
        assert isinstance(entity._geoasn, IpWhoisClient)
        assert isinstance(entity._whois, PTClient)
        assert entity._greynoise is None
        assert entity._urlhaus is None
        unset_env_vars()

    @patch("sys.argv", ["main", "1[.]1[.]1[.]1", "-s", "-g", "-u", "-a"])
    def test_handler_ip_2(self, fake_load_dotenv_1):
        """ IP with various options """
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_1):
            parse_env()
            console = Console()
            progress = simulate_progress(console),
            entity = generate_entity_handler(parse_args(), console, progress)
        assert isinstance(entity._geoasn, IpWhoisClient)
        assert isinstance(entity._whois, PTClient)
        assert isinstance(entity._shodan, ShodanClient)
        assert isinstance(entity._greynoise, GreynoiseClient)
        assert isinstance(entity._urlhaus, UrlHausClient)
        assert isinstance(entity._abuseipdb, AbuseIpDbClient)
        unset_env_vars()


class TestGenView:
    """ Tests for the generate_view function """
    @patch("wtfis.main.DomainView", return_value=MagicMock(spec=DomainView))
    def test_view_domain_1(self, m_domain_view, test_data):
        """ Domain view with default params """
        entity = DomainHandler(
            entity=MagicMock(),
            console=MagicMock(),
            progress=MagicMock(),
            vt_client=MagicMock(),
            ip_geoasn_client=MagicMock(),
            whois_client=MagicMock(),
            shodan_client=MagicMock(),
            greynoise_client=MagicMock(),
            abuseipdb_client=MagicMock(),
            urlhaus_client=MagicMock(),
        )
        entity.vt_info = Domain.model_validate(json.loads(test_data("vt_domain_gist.json")))
        entity.whois = MagicMock()
        entity.ip_enrich = MagicMock()
        view = generate_view(MagicMock(), MagicMock(), entity)
        assert isinstance(view, DomainView)

    @patch("wtfis.main.IpAddressView", return_value=MagicMock(spec=IpAddressView))
    def test_view_ip_1(self, m_ip_view, test_data):
        """ IP address view with default params """
        entity = IpAddressHandler(
            entity=MagicMock(),
            console=MagicMock(),
            progress=MagicMock(),
            vt_client=MagicMock(),
            ip_geoasn_client=MagicMock(),
            whois_client=MagicMock(),
            shodan_client=MagicMock(),
            greynoise_client=MagicMock(),
            abuseipdb_client=MagicMock(),
            urlhaus_client=MagicMock(),
        )
        entity.vt_info = IpAddress.model_validate(json.loads(test_data("vt_ip_1.1.1.1.json")))
        entity.whois = MagicMock()
        entity.ip_enrich = MagicMock()
        view = generate_view(MagicMock(), MagicMock(), entity)
        assert isinstance(view, IpAddressView)

    def test_view_error(self):
        """ IP address view with default params """
        with pytest.raises(Exception):
            generate_view(MagicMock(), MagicMock(), "foobar")


class TestMain:
    @patch("sys.argv", ["main", "www.example.com"])
    @patch("wtfis.main.Console", return_value=Console())
    @patch("wtfis.main.parse_args")
    @patch("wtfis.main.get_progress")
    @patch("wtfis.main.generate_entity_handler", return_value=MagicMock())
    @patch("wtfis.main.generate_view", return_value=MagicMock())
    def test_main_default(self, m_view, m_handler, m_progress, m_args, m_console, fake_load_dotenv_1):
        """ Test all calls with default values """
        m_args.return_value = parse_args()
        m_progress.return_value = simulate_progress(m_console())
        with patch("wtfis.main.load_dotenv", fake_load_dotenv_1):
            main()
        m_args.assert_called_once_with()
        m_handler.assert_called_once_with(m_args(), m_console(), m_progress())
        m_handler().fetch_data.assert_called_once_with()
        m_handler().print_warnings.assert_called_once_with()
        m_view.assert_called_once_with(m_args(), m_console(), m_handler())
        m_view().print.assert_called_once_with(one_column=False)
        unset_env_vars()



class Theme:
    panel_title = "bold yellow"
    heading_h1 = "bold bright_green on dark_green"
    heading_h2 = "bold yellow"
    table_field = "bold bright_magenta"
    table_value = "not bold default"
    popularity_source = "bright_cyan"
    inline_stat = "cyan"
    vendor_list = "cyan"
    nameserver_list = "bright_blue"
    disclaimer = "italic white on red"
    tags = "bright_white on black"
    tags_green = "bright_green on black"
    tags_red = "bright_red on black"
    product = "orange_red1"
    port = "bright_cyan"
    transport = "cyan"
    footer = "cyan"
    info = "bold green"
    warn = "bold yellow"
    error = "bold red"
    timestamp_date = "not bold default"
    timestamp_t = "dim bright_white"
    timestamp_time = "dim default"
    timestamp_z = "dim bright_white"
    asn_org = "bright_white"
    whois_org = "bright_cyan"
    urlhaus_bl_name = "bright_cyan"
    urlhaus_bl_low = "bright_white on green"
    urlhaus_bl_med = "black on yellow"
    urlhaus_bl_high = "bright_white on red"



"""
Logic handler for IP address inputs
"""
from wtfis.handlers.base import (
    BaseHandler,
    common_exception_handler,
    failopen_exception_handler,
)


class IpAddressHandler(BaseHandler):
    @common_exception_handler
    def _fetch_vt_ip_address(self) -> None:
        self.vt_info = self._vt.get_ip_address(self.entity)

    @common_exception_handler
    @failopen_exception_handler("_urlhaus")
    def _fetch_urlhaus(self) -> None:
        if self._urlhaus:
            self.urlhaus = self._urlhaus.enrich_ips(self.entity)

    def fetch_data(self):
        task_v = self.progress.add_task("Fetching data from Virustotal")
        self.progress.update(task_v, advance=50)
        self._fetch_vt_ip_address()
        self.progress.update(task_v, completed=100)

        task_g = self.progress.add_task(f"Fetching IP location and ASN from {self._geoasn.name}")
        self.progress.update(task_g, advance=50)
        self._fetch_geoasn(self.entity)
        self.progress.update(task_g, completed=100)

        if self._shodan:
            task_s = self.progress.add_task(f"Fetching IP data from {self._shodan.name}")
            self.progress.update(task_s, advance=50)
            self._fetch_shodan(self.entity)
            self.progress.update(task_s, completed=100)

        if self._urlhaus:
            task_u = self.progress.add_task(f"Fetching IP data from {self._urlhaus.name}")
            self.progress.update(task_u, advance=50)
            self._fetch_urlhaus()
            self.progress.update(task_u, completed=100)

        if self._greynoise:
            task_g = self.progress.add_task(f"Fetching IP data from {self._greynoise.name}")
            self.progress.update(task_g, advance=50)
            self._fetch_greynoise(self.entity)
            self.progress.update(task_g, completed=100)

        if self._abuseipdb:
            task_a = self.progress.add_task(f"Fetching IP data from {self._abuseipdb.name}")
            self.progress.update(task_a, advance=50)
            self._fetch_abuseipdb(self.entity)
            self.progress.update(task_a, completed=100)

        task_w = self.progress.add_task(f"Fetching IP whois from {self._whois.name}")
        self.progress.update(task_w, advance=50)
        self._fetch_whois()
        self.progress.update(task_w, completed=100)



from wtfis.clients.base import BaseRequestsClient, BaseWhoisClient
from wtfis.models.passivetotal import Whois
from wtfis.utils import refang


class PTClient(BaseRequestsClient, BaseWhoisClient):
    """
    Passivetotal client
    """
    baseurl = "https://api.riskiq.net/pt/v2"

    def __init__(self, api_user: str, api_key: str) -> None:
        super().__init__()
        self.s.auth = (api_user, api_key)

    @property
    def name(self) -> str:
        return "Passivetotal"

    def _query(self, path: str, query: str) -> dict:
        return self._get(
            path,
            params={"query": query}
        )

    def get_passive_dns(self, domain: str) -> dict:
        return self._query("/dns/passive", refang(domain))

    def get_whois(self, entity: str) -> Whois:
        return Whois.model_validate(self._query("/whois", refang(entity)))



import abc

from pydantic import ValidationError
from requests.exceptions import (
    ConnectionError,
    HTTPError,
    JSONDecodeError,
    RequestException,
    Timeout,
)
from rich.console import Console
from rich.progress import Progress
from shodan.exception import APIError
from typing import Callable, List, Optional, Union

from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient
from wtfis.models.abuseipdb import AbuseIpDbMap
from wtfis.models.base import WhoisBase
from wtfis.models.greynoise import GreynoiseIpMap
from wtfis.models.ipwhois import IpWhoisMap
from wtfis.models.shodan import ShodanIpMap
from wtfis.models.urlhaus import UrlHausMap
from wtfis.models.types import IpGeoAsnMapType
from wtfis.models.virustotal import Domain, IpAddress
from wtfis.clients.types import IpGeoAsnClientType, IpWhoisClientType
from wtfis.ui.theme import Theme
from wtfis.utils import error_and_exit, refang


def common_exception_handler(func: Callable) -> Callable:
    """ Decorator for handling common fetch errors """
    def inner(*args, **kwargs) -> None:
        progress: Progress = args[0].progress  # args[0] is the method's self input
        try:
            func(*args, **kwargs)
        except (APIError, ConnectionError, HTTPError, JSONDecodeError, Timeout) as e:
            progress.stop()
            error_and_exit(f"Error fetching data: {e}")
        except ValidationError as e:
            progress.stop()
            error_and_exit(f"Data model validation error: {e}")
    return inner


def failopen_exception_handler(client_attr_name: str) -> Callable:
    """ Decorator for handling calls that can fail open """
    def inner(func):
        def wrapper(*args, **kwargs) -> None:
            client = getattr(args[0], client_attr_name)  # Client obj who made the call
            warnings: List[str] = args[0].warnings
            try:
                func(*args, **kwargs)
            except (APIError, RequestException) as e:
                # Add warning
                warnings.append(f"Could not fetch {client.name}: {e}")
        return wrapper
    return inner


class BaseHandler(abc.ABC):
    def __init__(
        self,
        entity: str,
        console: Console,
        progress: Progress,
        vt_client: VTClient,
        ip_geoasn_client: IpGeoAsnClientType,
        whois_client: IpWhoisClientType,
        shodan_client: Optional[ShodanClient],
        greynoise_client: Optional[GreynoiseClient],
        abuseipdb_client: Optional[AbuseIpDbClient],
        urlhaus_client: Optional[UrlHausClient],
    ):
        # Process-specific
        self.entity = refang(entity)
        self.console = console
        self.progress = progress

        # Clients
        self._vt = vt_client
        self._geoasn = ip_geoasn_client
        self._whois = whois_client
        self._shodan = shodan_client
        self._greynoise = greynoise_client
        self._abuseipdb = abuseipdb_client
        self._urlhaus = urlhaus_client

        # Dataset containers
        self.vt_info: Union[Domain, IpAddress]
        self.geoasn: IpGeoAsnMapType = IpWhoisMap.empty()  # Default to ipwhois
        self.whois: WhoisBase
        self.shodan: ShodanIpMap = ShodanIpMap.empty()
        self.greynoise: GreynoiseIpMap = GreynoiseIpMap.empty()
        self.abuseipdb: AbuseIpDbMap = AbuseIpDbMap.empty()
        self.urlhaus: UrlHausMap = UrlHausMap.empty()

        # Warning messages container
        self.warnings: List[str] = []

    @abc.abstractmethod
    def fetch_data(self) -> None:
        """ Main method that controls what get fetched """
        return NotImplemented  # type: ignore  # pragma: no coverage

    @common_exception_handler
    @failopen_exception_handler("_geoasn")
    def _fetch_geoasn(self, *ips: str) -> None:
        self.geoasn = self._geoasn.enrich_ips(*ips)

    @common_exception_handler
    @failopen_exception_handler("_shodan")
    def _fetch_shodan(self, *ips: str) -> None:
        if self._shodan:
            self.shodan = self._shodan.enrich_ips(*ips)

    @common_exception_handler
    @failopen_exception_handler("_greynoise")
    def _fetch_greynoise(self, *ips: str) -> None:
        if self._greynoise:
            self.greynoise = self._greynoise.enrich_ips(*ips)

    @common_exception_handler
    @failopen_exception_handler("_abuseipdb")
    def _fetch_abuseipdb(self, *ips: str) -> None:
        if self._abuseipdb:
            self.abuseipdb = self._abuseipdb.enrich_ips(*ips)

    @common_exception_handler
    def _fetch_whois(self) -> None:
        # Let continue if rate limited
        try:
            self.whois = self._whois.get_whois(self.entity)
        except HTTPError as e:
            if e.response.status_code == 429:
                self.warnings.append(f"Could not fetch Whois: {e}")
            else:  # pragma: no coverage
                raise

    def print_warnings(self):
        for message in self.warnings:
            self.console.print(f"WARN: {message}", style=Theme().warn)



from pydantic import Field, model_validator
from typing import List, Optional

from wtfis.models.base import WhoisBase


class Whois(WhoisBase):
    source: str = "passivetotal"
    registrar: Optional[str] = None
    organization: Optional[str] = None
    name: Optional[str] = None
    email: Optional[str] = Field(None, alias="contactEmail")
    phone: Optional[str] = Field(None, alias="telephone")
    street: Optional[str] = None
    city: Optional[str] = None
    state: Optional[str] = None
    country: Optional[str] = None
    postal_code: Optional[str] = Field(None, alias="postalCode")
    name_servers: List[str] = Field([], alias="nameServers")
    date_created: Optional[str] = Field(None, alias="registered")
    date_changed: Optional[str] = Field(None, alias="registryUpdatedAt")
    date_expires: Optional[str] = Field(None, alias="expiresAt")
    dnssec: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def extract_registrant(cls, v):
        registrant = v.pop("registrant")
        for field in ["telephone", "street", "city", "state", "country", "postalCode"]:
            v[field] = registrant.get(field)
        return v



from __future__ import annotations

import abc
import sys

from pydantic import BaseModel, BeforeValidator, ConfigDict, RootModel
from pydantic.v1.validators import str_validator
from typing import List, Mapping, Optional

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated  # pragma: no coverage


LaxStr = Annotated[str, BeforeValidator(str_validator)]


class WhoisBase(BaseModel, abc.ABC):
    """ Use to normalize WHOIS fields from different sources """
    source: Optional[str] = None
    domain: Optional[str] = None
    registrar: Optional[str] = None
    organization: Optional[str] = None
    name: Optional[str] = None
    email: Optional[str] = None
    phone: Optional[str] = None
    street: Optional[str] = None
    city: Optional[str] = None
    state: Optional[str] = None
    country: Optional[str] = None
    postal_code: Optional[str] = None
    name_servers: List[str] = []
    date_created: Optional[str] = None
    date_changed: Optional[str] = None
    date_expires: Optional[str] = None
    dnssec: Optional[str] = None


class IpGeoAsnBase(BaseModel, abc.ABC):
    """ Use to normalize IP geolocation and ASN fields """
    model_config = ConfigDict(coerce_numbers_to_str=True)

    ip: str

    # Geolocation
    city: Optional[str]
    continent: Optional[str]
    country: Optional[str]
    region: Optional[str]

    # ASN
    asn: Optional[str]
    org: Optional[str]
    isp: Optional[str]
    domain: Optional[str]


class IpGeoAsnMapBase(RootModel, abc.ABC):
    root: Mapping[str, IpGeoAsnBase]

    @classmethod
    def empty(cls) -> IpGeoAsnMapBase:
        return cls.model_validate({})  # pragma: no coverage



"""
Type aliases
"""
from typing import Union

from wtfis.clients.ip2whois import Ip2WhoisClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.passivetotal import PTClient
from wtfis.clients.virustotal import VTClient


# IP geolocation and ASN client types
IpGeoAsnClientType = Union[
    IpWhoisClient,
]

# IP whois client types
IpWhoisClientType = Union[
    Ip2WhoisClient,
    PTClient,
    VTClient,
]



"""
Logic handler for domain and hostname inputs
"""
from typing import Optional

from requests.exceptions import HTTPError
from rich.console import Console
from rich.progress import Progress

from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient

from wtfis.handlers.base import (
    BaseHandler,
    common_exception_handler,
    failopen_exception_handler,
)

from wtfis.models.virustotal import Resolutions
from wtfis.clients.types import IpGeoAsnClientType, IpWhoisClientType


class DomainHandler(BaseHandler):
    def __init__(
        self,
        entity: str,
        console: Console,
        progress: Progress,
        vt_client: VTClient,
        ip_geoasn_client: IpGeoAsnClientType,
        whois_client: IpWhoisClientType,
        shodan_client: Optional[ShodanClient],
        greynoise_client: Optional[GreynoiseClient],
        abuseipdb_client: Optional[AbuseIpDbClient],
        urlhaus_client: Optional[UrlHausClient],
        max_resolutions: int = 0,
    ):
        super().__init__(entity, console, progress, vt_client, ip_geoasn_client, whois_client,
                         shodan_client, greynoise_client, abuseipdb_client, urlhaus_client)

        # Extended attributes
        self.max_resolutions = max_resolutions
        self.resolutions: Optional[Resolutions] = None

    @common_exception_handler
    def _fetch_vt_domain(self) -> None:
        self.vt_info = self._vt.get_domain(self.entity)

    @common_exception_handler
    def _fetch_vt_resolutions(self) -> None:
        # Let continue if rate limited
        try:
            self.resolutions = self._vt.get_domain_resolutions(self.entity)
        except HTTPError as e:
            if e.response.status_code == 429:
                self.warnings.append(f"Could not fetch Virustotal resolutions: {e}")
            else:
                raise

    @common_exception_handler
    @failopen_exception_handler("_urlhaus")
    def _fetch_urlhaus(self) -> None:
        if self._urlhaus:
            self.urlhaus = self._urlhaus.enrich_domains(self.entity)

    def fetch_data(self):
        task_v = self.progress.add_task("Fetching data from Virustotal")
        self.progress.update(task_v, advance=33)
        self._fetch_vt_domain()
        if self.max_resolutions != 0:
            self.progress.update(task_v, advance=33)
            self._fetch_vt_resolutions()
        self.progress.update(task_v, completed=100)

        if self.resolutions and self.resolutions.data:
            task_g = self.progress.add_task(f"Fetching IP location and ASN from {self._geoasn.name}")
            self.progress.update(task_g, advance=50)
            self._fetch_geoasn(*self.resolutions.ip_list(self.max_resolutions))
            self.progress.update(task_g, completed=100)

            if self._shodan:
                task_s = self.progress.add_task(f"Fetching IP data from {self._shodan.name}")
                self.progress.update(task_s, advance=50)
                self._fetch_shodan(*self.resolutions.ip_list(self.max_resolutions))
                self.progress.update(task_s, completed=100)

            if self._greynoise:
                task_g = self.progress.add_task(f"Fetching IP data from {self._greynoise.name}")
                self.progress.update(task_g, advance=50)
                self._fetch_greynoise(*self.resolutions.ip_list(self.max_resolutions))
                self.progress.update(task_g, completed=100)

            if self._abuseipdb:
                task_g = self.progress.add_task(f"Fetching IP data from {self._abuseipdb.name}")
                self.progress.update(task_g, advance=50)
                self._fetch_abuseipdb(*self.resolutions.ip_list(self.max_resolutions))
                self.progress.update(task_g, completed=100)

        if self._urlhaus:
            task_u = self.progress.add_task(f"Fetching domain data from {self._urlhaus.name}")
            self.progress.update(task_u, advance=50)
            self._fetch_urlhaus()
            self.progress.update(task_u, completed=100)

        task_w = self.progress.add_task(f"Fetching domain whois from {self._whois.name}")
        self.progress.update(task_w, advance=50)
        self._fetch_whois()
        self.progress.update(task_w, completed=100)



"""
ipwhois datamodels
API doc: https://ipwhois.io/documentation
"""
from __future__ import annotations

from typing import Dict

from pydantic import AliasPath, Field

from wtfis.models.base import IpGeoAsnBase, IpGeoAsnMapBase


class IpWhois(IpGeoAsnBase):
    # Metadata
    source: str = "IPWhois"

    # Results
    ip: str
    city: str
    region: str
    country: str
    continent: str
    asn: str = Field(validation_alias=AliasPath("connection", "asn"))
    org: str = Field(validation_alias=AliasPath("connection", "org"))
    isp: str = Field(validation_alias=AliasPath("connection", "isp"))
    domain: str = Field(validation_alias=AliasPath("connection", "domain"))


class IpWhoisMap(IpGeoAsnMapBase):
    root: Dict[str, IpWhois]

    @classmethod
    def empty(cls) -> IpWhoisMap:
        return cls.model_validate({})



from __future__ import annotations

from pydantic import BaseModel, RootModel, field_validator
from typing import Dict, List, Optional, Set


class Blacklists(BaseModel):
    spamhaus_dbl: str
    surbl: str


class Url(BaseModel):
    id: str
    urlhaus_reference: str
    url: str
    url_status: str
    date_added: str
    threat: str
    reporter: str
    larted: bool
    takedown_time_seconds: Optional[int]
    tags: List[str] = []

    @field_validator("larted", mode="before")
    @classmethod
    def convert_larted(cls, v):
        """ Cast larted to bool """
        mapping = {"true": True, "false": False}
        return mapping[v.lower()]

    @field_validator("tags", mode="before")
    @classmethod
    def handle_none_tags(cls, v):
        """ Turn NoneType tags into an empty list """
        return [] if v is None else v


class UrlHaus(BaseModel):
    query_status: str
    urlhaus_reference: Optional[str] = None
    host: Optional[str] = None
    firstseen: Optional[str] = None
    url_count: Optional[int] = None
    blacklists: Optional[Blacklists] = None
    urls: List[Url] = []
    _online_url_count: Optional[int] = None
    _tags: Set[str] = set()

    # Extracted fields

    @field_validator("url_count", mode="before")
    @classmethod
    def convert_url_count(cls, v):
        """ Cast url_count to int """
        return int(v) if v else None

    @property
    def online_url_count(self) -> int:
        if not self._online_url_count:
            self._online_url_count = (
                len([u for u in self.urls if u.url_status == "online"])
                if self.urls else 0
            )
        return self._online_url_count

    @property
    def tags(self) -> List[str]:
        if not self._tags:
            for url in self.urls:
                for tag in url.tags:
                    self._tags.add(tag)
        return sorted(self._tags)


class UrlHausMap(RootModel):
    root: Dict[str, UrlHaus]

    @classmethod
    def empty(cls) -> UrlHausMap:
        return cls.model_validate({})
